#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Dec 19 16:22:53 2020

@author: af1tang
"""
import torch, torch.nn.functional as F
from convogym.load_data import default_action_space as action_space
from convogym.utils._reshape import flatten
from convogym.utils._device import to_var
from convogym.utils._sampling import top_k_top_p_filtering

#### Baseline Agent ####
class Agent(object):
    """
    Wrapper object for the decoder model to do conditional decoding (no gradients). 
    
        - An Agent object can take on either the role of person 1 (user / policy) or person 2 (agent). 
        - The decoder model is assumed to be a transformer decoder (e.g., GPT2LMHeadModel), so top-k / nucleus sampling is used during token generation. 

    Parameters
    ----------
    model : A Huggingface transformer decoder model (default = GPT2LMHeadModel).
        Currently supports Pytorch Huggingface (transformers) models only. 
        The default model is af1tang/personaGPT.
        
    tokenizer : Huggingface transformer tokenizer (default = GPT2Tokenizer).
        The corresponding tokenizer for the transformer model. 
        The default tokenizer is af1tang/personaGPT.
        
    personas : List of strings
        List of persona facts to be included in the prefix tokens inputs to the decoder model.
        
    reverse : bool, optional
        If True, the Agent object takes on the role of person 1 (user / policy). The default is False.
    
    top_k : int, optional
        Number of top candidates to consider (sample from) the conditional distribution at each decoding step. The default is 10.
        
    top_p : float, optional
        Nucleus sampling parameter. Top fraction (0-1.0) of the conditional probability mass function to sample from at each decoding step. The default is .92.
        
    max_length : int, optional
        Max number of tokens allowed to decode in a given conversation. The default is 1024.

    Attributes
    -------
    p1 : List of strings
        List of persona facts corresponding to person 1.
        
    p2: List of strings
        List of persona facts corresponding to person 2.
        
    reversed : bool
        If True, the Agent object takes on the role of person 1 (user / policy). The default is False.
    
    inp : torch.LongTensor
        Tensor (1 x num tokens) of input tokens to the decoder model:
            - persona token <|p1|> or <|p2|> indicating the person identity OR
                action token <|act|> followed by the turn-level goal
            - persona facts separated by <|endoftext|> tokens
            - <|sep|> token separates persona facts or turn-level goal from dialog history
            - <|start|> token denotes start of dialog history
            - dialog history up to current turn.
        
        The input tokens can be thought of as the prompt used by the decoder (language model) to autoregressively decode the next set of tokens that comprise the response.
    
    past : torch.Tensor
        Cached hidden states of each layer in the transformer for quicker inference.
    
    turn : int 
        Turn count in the current conversation.
        
    curr_len : int
        Total number of tokens so far in the current conversation.
        
    dialog_history : List of list of ints
        List of token IDs for each response in the conversation.

    """
    def __init__(self, model, tokenizer, persona, reverse = False,
                 top_k=10, top_p = .92, max_length=1024):
        self.model = model
        self.tokenizer = tokenizer
        self.top_k, self.top_p, self.max_length = top_k, top_p, max_length
        self.reversed = reverse
        if not self.reversed:
            self.p1 = [] # self.NA_token # don't know your partner's persona
            self.p2 = persona
        else:
            self.p2 = [] #self.NA_token
            self.p1 = persona
        self.reset_convo()
        
    def __call__(self, inp, act=False):
        """
        Calls self.step(inp, act).

        Parameters
        ----------
           
        inp : List of ints
            Token IDs of the input message.
            
        act : bool, optional
            Whether the input prefix is an action code (True) or a set of persona facts (False). The default is False.

        Returns
        -------
        List of ints
            Token IDs of the output message.

        """
        return self.step( inp, act)
        
    def reset_convo(self):
        """
        Resets the dialog history and turn count. Used at beginning of conversations.
        """
        # reset dialog history
        self.dialog_history, self.turn = [], 0
        
    def _update_prefix(self, action):
        """
        Updates the prefix tokens with the input turn-level goal description.

        Parameters
        ----------
        action : List of strings
            A list with a single string representing the turn-level goal.

        Raises
        ------
        NotImplementedError
            Raises an error if the input action is not in the support set of actions.

        Returns
        -------
        None.

        """
        if action not in action_space:
            raise NotImplementedError("this action is not currently in the set of learned action space")
        else:
            if self.reversed:
                self.p2 = [action]
            else:
                self.p1 = [action]
    
    def _reset_inp(self, act = False):
        """
        Formats the input to the agent. If self.reversed=True, use person 1 facts, otherwise use person 2 facts. 
        
        If act=True, self.p1 or self.p2 represents the turn-level goal rather than a list of persona facts. 

        Parameters
        ----------
        act : bool, optional
            Whether the input prefix is an action code (True) or a set of persona facts (False). The default is False.

        Returns
        -------
        None.

        """
        if not act:
            if self.reversed:
                self.inp = self.tokenizer.encode(''.join(['<|p1|>'] + self.p1 + ['<|sep|>'] + ['<|start|>']))
            else:
                self.inp = self.tokenizer.encode(''.join(['<|p2|>'] + self.p2 + ['<|sep|>'] + ['<|start|>']))
        else:
            if self.reversed:
                self.inp =  self.tokenizer.encode(''.join(['<|act|> '] + self.p2 + ['<|sep|>'] + ['<|p1|>'] + self.p2 + ['<|sep|>'] + ['<|start|>']))
            else:
                self.inp =  self.tokenizer.encode(''.join(['<|act|> '] + self.p1 + ['<|sep|>'] + ['<|p1|>'] + self.p2 + ['<|sep|>'] + ['<|start|>']))
        # incorporate dialog dx
        self.inp += flatten(self.dialog_history)
        self.inp, self.curr_len, self.past = to_var(torch.tensor([self.inp])), len(self.inp), None

    def _update_dialog_hx(self, new_inp):
        """
        Updates the dialog history with new input message.

        Parameters
        ----------
        new_inp : List of ints
            List of tokens corresponding to the input message.

        Returns
        -------
        None.

        """
        if new_inp is not None:
            self.dialog_history.append(new_inp)
        
    def step(self, inp, act=False):
        """
        Generates a response message based on input message.

        Parameters
        ----------

        inp : List of ints
            Token IDs of the input message.
            
        act : bool, optional
            Whether the input prefix is an action code (True) or a set of persona facts (False). The default is False.

        Returns
        -------
        List of ints
            Token IDs of the output message.
            
        """
        self._update_dialog_hx(inp)
        self._reset_inp(act)
        outp = []
        self.model.eval()
        with torch.no_grad():
            while (self.tokenizer.eos_token_id not in outp) and (self.curr_len + len(outp) < self.max_length):
                outputs = self.model(self.inp, past_key_values=self.past)
                logits, self.past = outputs.logits, outputs.past_key_values
                # top k sampling          
                log_scores = top_k_top_p_filtering(logits[:,-1,:], top_k=self.top_k, top_p=self.top_p)
                probs = F.softmax(log_scores, dim=-1)
                token = torch.multinomial(probs, num_samples=1).squeeze(1)
                # update tokens for next output
                outp += token.tolist()
                self.inp = token.unsqueeze(0) 
                self.curr_len+=1
        self.dialog_history.append(outp)
        self.turn+=1
        return outp
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Aug 19 14:23:34 2021

@author: af1tang
"""

flatten = lambda l: [item for sublist in l for item in sublist]

def chunker(seq, size):
    """
    originally from
    https://stackoverflow.com/questions/434287/what-is-the-most-pythonic-way-to-iterate-over-a-list-in-chunks
    """
    return (seq[pos:pos + size] for pos in range(0, len(seq), size))

def split_by_index(seq, sep):
    """
    originally from 
    https://stackoverflow.com/questions/15357830/splitting-a-list-based-on-a-delimiter-word
    """
    result = []
    for el in seq:
        result.append(el)
        if el == sep:
            yield result
            result = []


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Oct 12 14:21:35 2020

@author: af1tang
"""
from dotenv import load_dotenv
import pkg_resources
import os, torch, pickle

load_dotenv(verbose=True)
# paths and configs
base_path = os.path.join(os.environ['HOME'], 'convogym/')
save_path = os.path.join(base_path, 'checkpoint')
model_path = os.path.join(base_path, 'checkpoint/model/')
data_path = pkg_resources.resource_filename('convogym', 'data/')
example_path = os.path.join(save_path, 'example/')


def create_dir(directory):
    """
    Create directory if not exists.
    """
    if not os.path.exists(directory):
        os.makedirs(directory)

# initialize save folder
create_dir(base_path)
create_dir(save_path)
create_dir(model_path)
create_dir(example_path)

class Configs():
    def __init__(self):
        # saving and loading paths
        self.model_path = model_path
        self.data_path = data_path
        self.save_path = save_path
        self.example_path = example_path
        self.plot_path = os.path.join(save_path,'samples/')
        self.download_name = 'af1tang/personaGPT'

opts = Configs()





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Sep  9 12:51:20 2021

@author: af1tang
"""
import time
import torch, torch.nn as nn, torch.nn.functional as F
from torch.utils.data import DataLoader

from torch.optim import AdamW
from transformers import get_linear_schedule_with_warmup
from convogym.utils._device import to_device
from convogym.utils._reshape import flatten
from convogym.utils._optim_helpers import fit_on_batch
from convogym.utils._visualization import plot_losses

def _get_grouped_params(model, act_tok=50262, start_tok=50259,
                                p1_tok=50260, p2_tok=50261):
    """
    Group learn rates by token groups:
        - new special tokens are "fast" 
        - regular tokens are frozen ("freeze", very slow rate)
        - positional weights are "slow"
        - everything else is "normal" 
        
    Note: this only applies if using af1tang/personaGPT model card.
    
    """
    with torch.no_grad():
        fast_group = flatten([[p[act_tok], p[start_tok], p[p1_tok], p[p2_tok]] for n,p in model.named_parameters() if n == 'transformer.wte.weight']) 
        freeze_group = [p[:start_tok] for n,p in model.named_parameters() if n == 'transformer.wte.weight']
        slow_group = [p for n,p in model.named_parameters() if n == 'transformer.wpe.weight']
        normal_group = [p for n,p in model.named_parameters() if n not in ('transformer.wte.weight',
                                                                           'transformer.wpe.weight')]
    
    optimizer_grouped_parameters = [{"params": fast_group, 'lr': 6.25e-4}, 
                                    {"params": freeze_group, 'lr': 1e-6}, 
                                    {"params": slow_group, 'lr': 1e-5}, 
                                    {"params": normal_group, 'lr': 5e-5}]
    return optimizer_grouped_parameters

class Learner(nn.Module):
    """
    Training wrapper for the decoder model. Updates the model using:  
        - active learning batch
        - supervised learning batch using given training_data.

    Parameters
    ----------
    model : A Huggingface transformer decoder model (default = GPT2LMHeadModel).
        Currently supports Pytorch Huggingface (transformers) models only. 
        The default model is af1tang/personaGPT model card.
        
    training_data : torch.utils.data.Dataset or List of tuples 
        List of training batches to sample batches from. Used to re-fit model to prevent catastrophic forgetting.
        
    lr : float, optional
        Learning rate for model parameters. The default is 5e-5.
        
    use_param_groups : bool, optional
        Whether to use different learn rates for special tokens, positional tokens and normal tokens.
        The default is True.
        
    schedule_func : torch.optim.lr_scheduler object, optional
        Learn rate scheduler function. The default is get_linear_schedule_with_warmup.
        
    gradient_accumulation_steps : int, optional
        Number of gradient accumulation steps. The default is 8.
        
    max_grad_norm : float, optional
        Max gradient norm size for gradient clipping. The default is 1.0.
        
    optim_func : torch.optim object, optional
        Optimizer used to update parameters. The default is AdamW.
        
    total_iters : int, optional
        Max number of training iters. The default is 20000.

    Attributes
    -------
    steps : int
        Tracks how many global update steps already taken during active learning.
    
    dataloader : torch.utils.data.DataLoader object
        Stores training_data as DataLoader object to generate batch samples.

    """
    def __init__(self, model, training_data, lr=5e-5, use_param_groups=False,
                 schedule_func=None,
                 gradient_accumulation_steps=8, max_grad_norm=1.0,
                 optim_func=None, total_iters=20000):
        super(Learner, self).__init__()
        if use_param_groups:
            optimizer_grouped_parameters = _get_grouped_params(model)
        else:
            optimizer_grouped_parameters = model.parameters()
        self.model = model
        if not optim_func:
            self.optimizer = AdamW(optimizer_grouped_parameters, lr=lr, eps=1e-8)
        if not schedule_func:
            self.scheduler = get_linear_schedule_with_warmup(self.optimizer, 
                                                num_warmup_steps=int(.2 * total_iters), 
                                                num_training_steps=total_iters)
        self.dataloader = DataLoader(training_data, batch_size=1, shuffle=True)
        self.data_iter = iter(self.dataloader)
        self.gradient_accumulation_steps = gradient_accumulation_steps
        self.max_grad_norm = max_grad_norm
        self.steps = 0
    
    def _reset_iterator(self):
        """
        Resets data iterator object using self.dataloader.
        
        Attributes
        -------
        data_iter : torch.utils.data.dataloader._SingleProcessDataLoaderIter
            Generator object to yield the next batch sample.

        Returns
        -------
        None.

        """
        self.data_iter = iter(self.dataloader)
        
    def forward(self, *args, **kwargs):
        """
        Calls self.model(*args, **kwargs), this part is the same as the original model.
        """
        return self.model(*args, **kwargs)
    
        
    def fit_on_active_batch(self, active_batch):
        """
        Updates model paramters based on input active learning batch.

        Parameters
        ----------
        active_batch : tuple
            Format of tuple is (x,y): 
                - X: List of int
                    Input tokens (encoded by tokenizer) of action prefix + dialog history.
                - Y: List of int
                    Encoded human-provided response (ground truth).
        Returns
        -------
        None.

        """
        print(); print("Fitting on batch ... ")
        self.model.train()
        new_loss = fit_on_batch(self.model, active_batch, self.gradient_accumulation_steps)
        try:
            old_batch = self.data_iter.next()
        except:
            self._reset_iterator()
            old_batch = self.data_iter.next()
        old_loss = fit_on_batch(self.model, old_batch, self.gradient_accumulation_steps)
        # step
        if (self.steps+1) % self.gradient_accumulation_steps == 0:
            torch.nn.utils.clip_grad_norm_(self.model.parameters(), self.max_grad_norm)
            self.optimizer.step()
            self.scheduler.step()  # Update learning rate schedule
            self.model.zero_grad()
            
            print('Iter: %d | new_loss: %.3f | old_loss: %.3f | lr: %s ' %( 
                    self.steps, new_loss.item() * self.gradient_accumulation_steps, 
                    old_loss.item() * self.gradient_accumulation_steps,
                    str(self.scheduler.get_last_lr()[0])) )
        self.steps += 1
        self.model.eval()
    
    def fit(self, num_train_epochs=3, logging_steps=10, save_steps=250):
        """
        Trains self.model on the provided dataset (self.dataloader). 
        

        Parameters
        ----------
        num_train_epochs : int, optional
            Total number of epochs . The default is 3.
            
        logging_steps : int, optional
            Training progress is printed every logging_steps. The default is 10.
            
        save_steps : int, optional
            Training loss and learn rate plots are saved every save_steps. The default is 250.

        Returns
        -------
        stats : dict
            Dictionary of training performance.
                - keys: global steps (self.steps)
                - values: pretrain loss and learn rate.

        """
        stats = {}        
        tr_loss, logging_loss = 0.0, 0.0
        # model to training mode
        self.model.zero_grad() 
        self.model.train()
        start_time = time.time()
        
        t_total = len(self.dataloader) // self.gradient_accumulation_steps * num_train_epochs
        for epoch in range(num_train_epochs):
            self._reset_iterator()
            for step in range(len(self.dataloader)):
                ### step ###
                batch = self.data_iter.next()
                loss = fit_on_batch(self.model, batch, self.gradient_accumulation_steps); del batch
                # logging (new data only)
                tr_loss += loss.item()
                
                # gradient accumulation
                if (step+1) % self.gradient_accumulation_steps == 0:
                    torch.nn.utils.clip_grad_norm_(self.model.parameters(), self.max_grad_norm)
                    self.optimizer.step()
                    self.scheduler.step()  # Update learning rate schedule
                    self.model.zero_grad()
                    self.steps += 1
                    
                    # reporting 
                    if (self.steps+1) % logging_steps ==0:
                        stats[self.steps] = {'pretrain_loss': (tr_loss - logging_loss) / logging_steps, 
                                              'pretrain_lr': self.scheduler.get_last_lr()[-1]}
                        logging_loss = tr_loss
                        
                        elapsed_time = time.strftime("%M:%S", time.gmtime(time.time() - start_time))
                        print('Epoch: %d | Iter: [%d/%d] | loss: %.3f | lr: %s | time: %s' %( 
                        epoch, self.steps, t_total, stats[self.steps]['pretrain_loss'],                             
                                str(stats[self.steps]['pretrain_lr']), elapsed_time))
                        start_time = time.time()
                        
                    if (self.steps + 1) % save_steps==0:
                        print("Plotting training loss and lr ... ")
                        plot_losses(stats, title='pretrain_loss' )
                        plot_losses(stats, title='pretrain_lr')
                        print("Done.")
        self.model.eval()
        return stats
        
    def save(self, save_path):
        """
        Saves model parameters and config to save_path.
        """
        self.model.save_pretrained(save_path)

    def load(self, load_path):
        """
        Tries to load model parameters from load_path, raises exception if fails.
        """
        try: 
            self.model.from_pretrained(load_path)
        except Exception as e:
            raise e


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Oct 12 14:17:28 2020

@author: af1tang
"""
import torch
device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

def to_device(model):
    return model.to(device)

def to_data(x):
    if torch.cuda.is_available():
        x = x.cpu()
    return x.data.numpy()

def to_var(x):
    if not torch.is_tensor(x):
        x = torch.Tensor(x)
    if torch.cuda.is_available():
        x = x.cuda()
    return x

            



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Aug 19 14:18:32 2021

@author: af1tang
"""
import torch, torch.nn.functional as F

## nucleus and top k sampling ##
def top_k_top_p_filtering(logits: torch.Tensor, top_k: int = 0,  top_p: float = 1.0, filter_value: float = -float("Inf"),
                          min_tokens_to_keep: int = 1) -> torch.Tensor:
    '''
    adapted from https://github.com/huggingface/transformers/blob/master/src/transformers/generation_utils.py
    '''
    if top_k > 0:
        top_k = min(max(top_k, min_tokens_to_keep), logits.size(-1))  # Safety check
        # Remove all tokens with a probability less than the last token of the top-k
        indices_to_remove = logits < torch.topk(logits, top_k)[0][..., -1, None]
        logits[indices_to_remove] = filter_value

    if top_p < 1.0:
        sorted_logits, sorted_indices = torch.sort(logits, descending=True)
        cumulative_probs = torch.cumsum(F.softmax(sorted_logits, dim=-1), dim=-1)

        # Remove tokens with cumulative probability above the threshold (token with 0 are kept)
        sorted_indices_to_remove = cumulative_probs > top_p
        if min_tokens_to_keep > 1:
            # Keep at least min_tokens_to_keep (set to min_tokens_to_keep-1 because we add the first one below)
            sorted_indices_to_remove[..., :min_tokens_to_keep] = 0
        # Shift the indices to the right to keep also the first token above the threshold
        sorted_indices_to_remove[..., 1:] = sorted_indices_to_remove[..., :-1].clone()
        sorted_indices_to_remove[..., 0] = 0

        # scatter sorted tensors to original indexing
        indices_to_remove = sorted_indices_to_remove.scatter(1, sorted_indices, sorted_indices_to_remove)
        logits[indices_to_remove] = filter_value
    return logits


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Sep 15 01:58:57 2021

@author: af1tang
"""

import pytest
from convogym import decoders
from convogym.utils._device import to_var

def test_loading_model_tokenizer():
    assert len(decoders.tokenizer) == 50263
    assert ( decoders.act_tok == 50262 ) and ( decoders.start_tok == 50259)
    assert ( decoders.p1_tok == 50260 ) and ( decoders.p2_tok == 50261)
    
def test_default_tokenizer():
    goal = "ask about hobbies."
    inp = decoders.tokenizer.encode("<|act|>" + goal + "<|p1|><|sep|><|start|>")        
    assert inp == [50262, 2093, 546, 45578, 13, 50260, 50257, 50259]
    
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Aug 19 16:00:29 2021

@author: af1tang
"""
import torch
from convogym.utils._device import to_var

class Env:        
    """
    A helper object for callbacks to estimate the state information at each turn during a dialog. 
    
    Parameters
    -------
    scb : StateCb object
        A callback object that tracks the states, contexts, actions objects that are updated during conversation.
        
    mcb : MessageCb object
        A callback object that tracks raw text, turn count and status of the conversation.
        
    state_estimator : object or function
        A custom state estimation object or function that maps from dialog history tokens (or token embeddings) to state vectors.
    
    state_size : int, optional
        Number of dimensions in the state vector.
    """
    def __init__(self, state_estimator, state_size = 1024):
        self.state_estimator = state_estimator
        self.state_size = state_size
        
    def reset(self, scb):
        """
        Initializes the state and context vectors to the 0 vectors. 
        """
        scb.state = to_var(torch.zeros(self.state_size))
        scb.states, scb.rewards, scb.actions = [scb.state.tolist() + [0.0]], [], []
        return scb
    
    def get_policy_inp(self, scb):
        """
        Formats the turn count and state into a single vector input to the policy.
        """
        turn_tensor = to_var(torch.ones(1,1)) * scb.turn
        state = scb.state.view(1,-1)
        state_t = torch.cat((state, turn_tensor), dim=-1)
        return state_t
    
    def get_curr_state(self, scb, mcb):
        """
        Use state estimator to obtain state and context vectors from dialog history text.
        
        Notes
        -----
        Only the person 2 responses are used to obtain the state and context vectors corresponding to person 2.
        """
        with torch.no_grad():
            scb.state = self.state_estimator(dialog_history=mcb.dialog_hx[1::2])
        return scb, mcb
        
    def get_next_state(self, scb, mcb):
        """
        Not implemented. Passes new state estimates to state callbacks.
        """
        _state = scb.state.tolist()
        _state.append(scb.turn+1.0)
        scb.states.append(_state)
        return scb, mcb


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Aug 21 14:34:48 2021

@author: af1tang
"""
import os, pickle, ast
import pandas as pd
import torch
from tqdm import tqdm
from convogym._configs import opts
from convogym.utils._reshape import flatten
from convogym.utils._device import to_var, to_data, device
from convogym.utils._visualization import to_tokens

default_action_space = ['ask about kids.',
                         'ask about pets.',
                         'talk about work.',
                         'ask about marital status.',
                         'talk about travel.',
                         'ask about age and gender.',
                         'ask about hobbies.',
                         'ask about favorite food.',
                         'talk about movies.',
                         'talk about music.',
                         'talk about politics.']

# methods for loading saved datasets 
def load_selfplay_convos(tokenizer, data_path):
    """
    Loads saved self-play convos into training data format.
    
    Parameters
    -------
    tokenizer : A transformer tokenizer e.g., GPT2Tokenizer
        Associated tokenizer for the transformer. 
        (default = af1tang/personaGPT)    
        
    data_path : String or os.path
        path to pandas dataframe of history (hx) and personas (p1 and p2)
        
    Results
    -------
    data : List of tuples
        - list of rows in df, where hx is represented as token_ids, 
        - p1 and p2 are list of strings. 
            
    """
    df = pd.read_csv(data_path)
    data = []
    for i, row in tqdm(df.iterrows(), total=len(df)):
        convo, p1, p2 = row['hx'], row['p1'], row['p2']
        convo, p1, p2 = ast.literal_eval(convo), ast.literal_eval(p1), ast.literal_eval(p2)

        x = [tokenizer.encode(line) for line in convo]
        data.append( [x, p1, p2] )
    return data

def load_active_learning_convos(tokenizer, data_path):
    """
    Loads saved active learning convos into training data format.

    Parameters
    ----------
    tokenizer : A transformer tokenizer e.g., GPT2Tokenizer
        Associated tokenizer for the transformer. 
        (default = af1tang/personaGPT)    
        
    data_path : String or os.path
        path to pandas dataframe of input context (X), response targets (y).
            
    Returns
    -------
    active_data : List of tuples
        List of training batches of the form (X,y) 
        where X is the prefix (turn-level goals + dialog history) 
        and y is the response tokens to decode.

    """
    df = pd.read_csv(data_path)
    X, y = df['X'].tolist(), df['y'].tolist()
    # active learning data
    active_data = list(zip(X,y))
    return active_data

def load_imitation_learning_convos(tokenizer, data_path):
    """
    Loads saved active learning convos into imitation learning format.

    Parameters
    ----------
    tokenizer : A transformer tokenizer e.g., GPT2Tokenizer
        Associated tokenizer for the transformer. 
        (default = af1tang/personaGPT)    
        
    data_path : String or os.path
        path to pandas dataframe of history (dialog_hx) and turn-level goals (actions).
            
    Returns
    -------
    imitation_data : List of tuples
        List of training batches of the form (dialog_hx, actions) 
        where dialog_hx is the tokens from which to extract state estimates,
        actions is the output actions taken by the human user. 
        
        One can use this data for imitation learning or to pretrain the policy.

    """
    df = pd.read_csv(data_path)
    dialog_history = df['dialog_hx'].tolist()
    actions = df['actions'].tolist()
    # imitation learning data
    imitation_data = list(zip(dialog_history, actions))
    return imitation_data

def load_rl_convos(tokenizer, state_estimator, data_path):
    """
    Loads saved RL conversations from convogym into training data format.

    Parameters
    ----------
    tokenizer : A transformer tokenizer e.g., GPT2Tokenizer
        Associated tokenizer for the transformer. 
        (default = af1tang/personaGPT)    
        
    state_estimator : StateEstimator object or Callable
        A state estimator model that maps from dialog history text -> embedding vector to represent state information.
        
    data_path : String or os.path
        path to pandas dataframe of input context (X), response targets (y), 
            history (dialog_hx), and turn-level goals (actions).
            
    Returns
    -------
    active_data : List of tuples
        List of training batches of the form (X,y) 
        where X is the prefix (turn-level goals + dialog history) 
        and y is the response tokens to decode.

    Returns
    -------
    mdp_data : List of tuples
        List of training batches of the form (dialog_hx, actions) 
        where dialog_hx is the tokens from which to extract state estimates,
        actions is the output actions taken by the human user. 

    """
    df = pd.read_csv(data_path)
    hx, actions = df['dialog_hx'].tolist(), df['actions'].tolist()
    rewards, personas = df['rewards'].tolist(), df['personas'].tolist()
    data = []
    print("building dataset ... ")
    for i, (convo, ep_actions, ep_rewards) in tqdm(list(zip(hx, actions, rewards)), total=len(df)):
        states, acts, rewards = [], [], [], []
        for turn in range(0,len(convo),2):
            curr_hx = convo[0:turn]
            if len(curr_hx) == 0:
                state = [0.0]*1025
            else:
                state = state_estimator(to_tokens(curr_hx[1::2], tokenizer)).tolist()
                state.append(turn/2.0)
                # get reward
                rewards.append(ep_rewards[turn//2])
            if turn // 2 < len(convo)/2:
                a = ep_actions[turn //2 ]; acts.append(a)
            states.append(state)

        # batching
        next_states = states[1:]; states = states[:-1]
        next_acts = acts[1:] + [0]
        batch = list(zip(states, acts, next_states, next_acts, rewards))
        data.append(batch)

    print("done.")
    return data

def load_persona_train(path=os.path.join(opts.example_path,'train_decoder_data'),
                       data_path=None):
    """
    Loads persona training set batches for fine-tuning (used during active learning).
    
    Parameters
    ----------
    path : os.path or string, optional
        Path to pickled file if exists. The default is os.path.join(opts.example_path,'train_decoder_data').

    data_path : os.path or string, optional
        Path to persona dataset. The default is convogym/data/personachat.csv.

    Returns
    -------
    train_data : List of tuples
        List of training batches.
    """
    try:
        with open(path, 'rb') as f:
            train_data = pickle.load(f)
    except Exception as e:
        print(e)
        print("Building PersonaChat dataset from scratch ... ")
        from convogym.decoders import model, tokenizer
        train_data, _ = prepare_personachat_dataset(model, tokenizer, 
                                                    data_path=data_path)
    return train_data

def load_state_estim_train(path=os.path.join(opts.example_path,'train_state_estim_data'),
                          data_path=None):
    """
    Loads PersonaChat and formats dialogs into state estimation format for training.    

    Parameters
    ----------
    path : os.path or string, optional
        Path to pickled file if exists. The default is os.path.join(opts.example_path,'train_state_estim_data').

    data_path : os.path or string, optional
        Path to persona dataset. The default is convogym/data/personachat.csv.

    Returns
    -------
    state_estim_data : List of tuples
        List of training batches.

    """
    try:
        with open(path, 'rb') as f:
            train_data = pickle.load(f)
    except Exception as e:
        print(e)
        print("Building state estimation data from PersonaChat ... ")
        from convogym.decoders import model, tokenizer
        train_data, _ = prepare_state_estim_dataset(model, tokenizer, 
                                                          data_path=None,
                                                          return_data=True)
    return train_data

# methods for building PersonaChat dataset 
def _load_personachat(tokenizer, data_path):
    """
    converts dataframe -> array of [tokens, p1, p2] 
    
    Parameters
    -------
    tokenizer : A transformer tokenizer e.g., GPT2Tokenizer
        Associated tokenizer for the transformer. 
        (default = af1tang/personaGPT)    
        
    data_path : String or os.path
        path to pandas dataframe of history (hx), personas (p1 and p2), 
            and split (train or test) as columns
    Results
    -------
    tr_data : List of tuples
        - list of rows in df, where hx is represented as token_ids, 
        - p1 and p2 are list of strings. 
        - selected rows where split == train.
            
    te_data: List of tuples
        - selected rows where split != train
    """
    df = pd.read_csv(data_path)
    tr_data, te_data = [], []
    for i, row in tqdm(df.iterrows(), total=len(df)):
        convo, p1, p2, split = row['hx'], row['p1'], row['p2'], row['split']
        convo, p1, p2 = ast.literal_eval(convo), ast.literal_eval(p1), ast.literal_eval(p2)

        x = [tokenizer.encode(line) for line in convo]
        if split == 'train':
            tr_data.append( [x, p1,p2])
        else:
            te_data.append( [x,p1,p2] )
    return tr_data, te_data


def _filter_personas(personas, uniques=None):
    """
    filters out redundant profiles from a list of personas.

    Parameters
    ----------
    personas : List of list of string
        List of persona profiles, each profile consists of 3-5 list of strings.
        
    uniques : List of string, optional
        Set of unique persona profiles. The default is None.

    Returns
    -------
    uniques : List of string, optional
        Set of unique persona profiles. The default is None..
        
    filtered : List of list of string, optional
        original list of personas, with each profile replaced by filtered personas.

    Examples
    -------
    >>> personas = [['i like wine', 'i have 5 kids'], 
                        ['i hate wine', 'i like ice cream'],
                        ['i like wine', 'i've 5 kids]] 
    >>> uniques, filtered = _filtered_personas) 
    >>> print(filtered)
    [[['i like wine', 'i have 5 kids'], 
      ['i hate wine', 'i like ice cream'], 
      ['i like wine', 'i have 5 kids']]
     
    """
    if not uniques: 
        uniques = []
    filtered = []
    
    for p in tqdm(personas, total=len(personas)):
        done, found = False, None
        i = 0
        while (not done) and len(uniques) > 0:
            p_cand_facts = set(uniques[i])
            num_overlap = sum([1 for fact in p if fact in p_cand_facts])
            if num_overlap > 1:
                done = True
                found = uniques[i]
            elif i+1 == len(uniques):
                done = True
            else:
                i += 1
        if found:
            filtered.append(found)
        else:
            uniques.append(p)
            filtered.append(p)
    print()
    print("%d unique persona profiles, %d unique persona facts found" %(len(uniques), 
                                                len(set(flatten(uniques)))
                                                                        )
          )
        
    return uniques, filtered

def _build_persona_batches(model, tokenizer, data):
    """
    Formats (dialog history, person 1 facts, person 2 facts) into turn-level batches for training.

    Parameters
    ----------
    model : A transformer decoder model e.g., GP2LMHeadModel 
        Transformer model (BERT, GPT, etc.) that embeds the tokens corresponding to the relevant turns in the dialog history. 
        (default = af1tang/personaGPT).
        
    tokenizer : A transformer tokenizer e.g., GPT2Tokenizer
        Associated tokenizer for the transformer. 
        (default = af1tang/personaGPT)        
        
    data : List of tuples
        List of training batches in (tokens, persona 1, persona 2) format, where persona 1 and persona 2 are list of strings.

    Returns
    -------
    result : List of tuples
        Each sample is a (x,y) pair:
            - x: token IDs of persona facts + dialog history up to current turn
            - y: token IDs of response at current turn.
        
        A dialog history of 8 turns (1 response from person 1 + 1 response from person 2 per turn) is broken up into 16 input output pairs. 
        
    Examples
    -------
    Suppose we have the following convo (a tuple) from PersonaChat:
    >>> dialog_tokens = ["hi how is it going?", # pre-tokenized
                         "not bad, how are you?",
                         "i'm having fun on fifa, you?",
                         "i'm tired from farming."]
    >>> dialog_hx = [tokenizer.encode(line) for line in dialog_tokens]
    >>> print(dialog_hx)
    [[5303, 703, 318, 340, 1016, 30],
     [1662, 2089, 11, 703, 389, 345, 30],
     [72, 1101, 1719, 1257, 319, 5515, 64, 11, 345, 30],
     [72, 1101, 10032, 422, 16035, 13]]
    >>> p1 = ["i like to play video games.", 
              "i want a cat someday."]
    >>> p2 = ["i work on a large farm.",
              "i don't have health insurance"]
    
    This convo gets converted to the following input-output pairs (x,y):
    >>> for i, (x,y) in enumerate(convo):
    >>>     print('-'*10+'turn %d'%i + '-' *10)
    >>>     print('input (x): ' , tokenizer.decode(x))
    >>>     print('output (y): ', tokenizer.decode(y))
    >>>     print()
    ----------turn 0----------
    input (x):  <|p1|> i like to play video games.i want a cat someday. <|sep|>
    output (y):  <|start|> hi how is it going?
    
    ----------turn 1----------
    input (x):  <|p2|> i work on a large farm.i don't have health insurance <|sep|> <|start|> hi how is it going?
    output (y):  not bad, how are you?
    
    ----------turn 2----------
    input (x):  <|p1|> i like to play video games.i want a cat someday. <|sep|> <|start|> hi how is it going?not bad, how are you?
    output (y):  i'm having fun on fifa, you?
    
    ----------turn 3----------
    input (x):  <|p2|> i work on a large farm.i don't have health insurance <|sep|> <|start|> hi how is it going?not bad, how are you?i'm having fun on fifa, you?
    output (y):  i'm tired from farming.        
    """
    result = []
    for (dialog_hx, p1, p2) in data:
        dialog_hx = dialog_hx[:20] # limit by max len of convo
        p1_ctx = tokenizer.encode(''.join(['<|p1|>'] + p1 + ['<|sep|>'] + ['<|start|>']))
        p2_ctx = tokenizer.encode(''.join(['<|p2|>'] + p2 + ['<|sep|>'] + ['<|start|>']))
        for t in range(len(dialog_hx)):
            x = dialog_hx[:t]
            y = dialog_hx[t]
            if t == 0:
                x = p1_ctx[:-1] 
                y = [p1_ctx[-1]] + y
            elif t %2 ==0:
                x = p1_ctx + flatten(x)
            else:
                x = p2_ctx + flatten(x)
            result.append((x,y))
    return result

def prepare_personachat_dataset(model, tokenizer, data_path=None,
                                save_dir=opts.example_path):
    """
    Converts persona dataset into language model task format.

    Parameters
    ----------
    model : A transformer decoder model e.g., GP2LMHeadModel 
        Transformer model (BERT, GPT, etc.) that embeds the tokens corresponding to the relevant turns in the dialog history. 
        (default = af1tang/personaGPT).
        
    tokenizer : A transformer tokenizer e.g., GPT2Tokenizer
        Associated tokenizer for the transformer. 
        (default = af1tang/personaGPT)   
        
    data_path : String or os.path
        File path for personachat data.

    Returns
    -------
    tr_batches : List of tuples
        List of training set samples.
        Each sample is a (x,y) pair:
            - x: token IDs of persona facts + dialog history up to current turn
            - y: token IDs of response at current turn.
        
        A dialog history of 8 turns (1 response from person 1 + 1 response from person 2 per turn) is broken up into 16 input output pairs. 
    
    te_batches : List of tuples
        List of test set samples.
        Each sample is a (x,y) pair:
            - x: token IDs of persona facts + dialog history up to current turn
            - y: token IDs of response at current turn.
        
        A dialog history of 8 turns (1 response from person 1 + 1 response from person 2 per turn) is broken up into 16 input output pairs. 
    
    """
    print()
    print("*"*50)
    print("Extracting data from personachat dataframe...")
    if not data_path:
        data_path = os.path.join(opts.data_path, 'personachat.csv')
    tr_data, te_data = _load_personachat(tokenizer, data_path)
    # build training and testing batches
    print("Building traing and testing batches from conversations...")
    tr_batches = _build_persona_batches(model, tokenizer, tr_data)
    te_batches = _build_persona_batches(model, tokenizer, te_data)
    # saving
    with open(os.path.join(save_dir, 'train_decoder_data'), 'wb') as f:
        pickle.dump(tr_batches, f)
    with open(os.path.join(save_dir, 'test_decoder_data'), 'wb') as f:
        pickle.dump(te_batches, f)
    print("done!")
    print()
    return tr_batches, te_batches

def prepare_state_estim_dataset(model, tokenizer, data_path=None,
                               save_dir=opts.example_path,
                               return_data=False):
    """
    Formats personachat dataframe into an array of training samples for unsupervised learning. 
    
    In unsupervised learning, embedding functions are learned for
        - the dialog history
        - the persona facts.
        
    Embeddings of the dialog history and persona facts can be used for state and context representations.


    Parameters
    ----------
    model : A transformer decoder model e.g., GP2LMHeadModel 
        Transformer model (BERT, GPT, etc.) that embeds the tokens corresponding to the relevant turns in the dialog history. 
        (default = af1tang/personaGPT).
        
    tokenizer : A transformer tokenizer e.g., GPT2Tokenizer
        Associated tokenizer for the transformer. 
        (default = af1tang/personaGPT)   
        
    data_path : String or os.path
        File path for personachat data.
    
    Saves
    -------
    tr_data : List of tuples
        List of training set samples. 
        Each sample is a tuple (dialog history, persona 1, persona 2):
            - dialog history : list of token ids for each conversational turn
            - persona 1 : list of strings of person 1 persona facts
            - persona 2 : list of strings of person 2 persona facts.
        
    te_data : List of tuples
        List of training set samples. 
        Each sample is a tuple (dialog history, persona 1, persona 2):
            - dialog history : list of token ids for each conversational turn
            - persona 1 : list of strings of person 1 persona facts
            - persona 2 : list of strings of person 2 persona facts.

    Returns
    -------
    persona_facts : List of strings
        List of unique persona facts in personachat.
    """
    print()
    print("*"*50)
    print("Extracting data from personachat dataframe...")
    if not data_path:
        data_path = os.path.join(opts.data_path, 'personachat.csv')
    tr_data, te_data = _load_personachat(tokenizer, data_path)
    # build filter personas to get unique persona facts
    x_tr, tr_p1, tr_p2 = list(zip(*tr_data))
    x_te, te_p1, te_p2 = list(zip(*te_data))
    # filter out redundant personas
    tr_uniques, tr_p1_filtered = _filter_personas(tr_p1)
    tr_uniques, tr_p2_filtered = _filter_personas(tr_p2, tr_uniques)
    te_uniques, te_p1_filtered = _filter_personas(te_p1)
    te_uniques, te_p2_filtered = _filter_personas(te_p2, te_uniques)
    persona_facts = list(set(flatten(tr_uniques + te_uniques)))
    # remake datasets w/ filtered personas 
    tr_data = list(zip(x_tr, tr_p1_filtered, tr_p2_filtered))
    te_data = list(zip(x_te, te_p1_filtered, te_p2_filtered))
    # saving
    df_facts = pd.DataFrame(persona_facts, columns=['Facts'])
    df_facts.to_csv(os.path.join(save_dir, 'persona_facts.csv'), index=False)
    with open(os.path.join(save_dir, 'train_state_estim_data'), 'wb') as f:
        pickle.dump(tr_data, f)
    with open(os.path.join(save_dir, 'test_state_estim_data'), 'wb') as f:
        pickle.dump(te_data, f)
    if not return_data:
        return persona_facts
    else:
        return tr_data, te_data
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Sep 15 13:05:11 2021

@author: af1tang
"""

import pytest
from convogym.policies import Policy # default policy
from convogym.utils._device import to_var
import torch, torch.nn as nn


def test_policy():
    action_space = [ 
        'talk about work.', 
        'ask about hobbies.',
        'talk about movies.'
        ]
    policy_net = nn.Sequential(nn.Linear(100, len(action_space))
                                )
    policy = Policy( policy = policy_net,
                        action_space = action_space) 
    action = policy.act( inp=to_var(torch.rand((1,100))), turn=1 ) 
    assert action in range(len(action_space))
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Aug 13 00:21:08 2021

@author: af1tang
"""

## state and dialog trackers 
class StateCb:
    """
    Callback object for tracking conversational state information such as (state, action, reward, next_state) at each turn of the conversation.

    Parameters
    ----------
    state : torch.Tensor, optional
        The feature vector embedding of the dialog history. Shape should be (1 x n) for a vector in R^n.
        In Markov Decision Process terms, this is the state input to the policy. The default is None.
        
        
    personas : List of strings, optional
        List of persona facts to track for scoring (if applicable). The default is None.
        
    actions : List of ints, optional
        List of action indices corresponding to actions taken per turn. The default is None.
        
    reward : float, optional
        Reward at a given turn. The default is None.
        
    act : string, optional
        Turn-level goal at a given turn in text form. The default is False.

    Attributes
    -------
    turn : int
        Tracks number of turns elapsed in current conversation.        
    
    done : bool
        Whether to end conversation. 
        
    rewards : List of floats
        Tracks reward received at each turn.

    """
    def __init__(self, state=None, personas = None,
                 actions=None, reward=None, act=False):

        self.state, self.action, self.reward, self.act = state, actions, reward, act
        self.states, self.actions, self.rewards = [], [], []
        self.personas = personas
        self.turn, self.done = 0, False

class MessageCb:
    """
    Tracks text information that flows throughout the conversation between the user and agent.

    Parameters
    ----------
    msg : List of ints, optional
        Input string message (tokenized) to a list of integers. The default is None.
        
    x : torch.LongTensor, optional
        Tracks the formatted prefix tokens (context / prompt) to the decoder. Used during ActiveGym to format the training data inputs. The default is None.
        
    y : torch.LongTensor, optional
        Tracks the formatted output tokens provided by human users. Used during ActiveGym to format the training data labels. The default is None.

    Attributes
    -------
    dialog_hx : List of list of ints
        List of token ids, each corresponding to a dialog response.
    """
    def __init__(self, msg=None, x=None, y=None):
        self.msg, self.x, self.y, self.dialog_hx = msg, x, y, []


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Sep 15 12:42:59 2021

@author: af1tang
"""

import pytest
from convogym.decoders import model, tokenizer
from convogym.gyms import Gym
from convogym.prefixes import get_custom_persona, get_random_persona


def test_self_play():
    gym = Gym(model=model, tokenizer=tokenizer, interactive=False, 
              reset_persona_func=get_random_persona, length=4)
    gym.sim_convos(num_convos=3)
    assert len(gym.data['hx']) == 3
    assert sum([ len(gym.data['hx'][i]) for i in range(3)]) /2 ==12
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Aug 19 16:02:24 2021

@author: af1tang
"""
import random
import numpy as np
import torch, torch.nn as nn
from convogym.utils._visualization import display_dialog_history
from convogym.prefixes import load_persona_facts

class RankingLoss(nn.Module):
    """
    Implements triplet loss with negative sampling.
    
    Parameters
    -------
    state_estimator : StateEstimator object or Callable
        A state estimator model that maps from dialog history text -> embedding vector to represent state information.

    swaps : bool, optional
        Whether to swap anchor and positive sample roles during triplet loss.
        
    margin : float, optional
        Margin parameter used for triplet loss.
    
    k : int, optional
        Number of total candidates to use during negative sampling. The default is 20.
        
    reduction : string, optional
        See https://pytorch.org/docs/stable/generated/torch.nn.TripletMarginLoss.html for details.
        > Specifies the reduction to apply to the output:
        >    ``'none'`` | ``'mean'`` | ``'sum'``. ``'none'``: no reduction will be applied,
        >    ``'mean'``: the sum of the output will be divided by the number of
        >    elements in the output, ``'sum'``: the output will be summed. Note: :attr:`size_average`
        >    and :attr:`reduce` are in the process of being deprecated, and in the meantime,
        >    specifying either of those two args will override :attr:`reduction`. Default: ``'mean'``
    """
    def __init__(self, state_estimator, swap=False, margin=1.0, k = 20, reduction='sum'):
        super(RankingLoss, self).__init__()
        self.ranking_loss = nn.TripletMarginLoss(margin=margin, swap=swap, reduction=reduction)
        self.state_estimator = state_estimator
        self.k = k
    
    def _calculate_loss(self, inp, pos_cands, neg_cands):
        """
        Calculates the triplet loss for a single sample.

        Parameters
        ----------
        inp : torch.Tensor
            A single state embedding.
            
        pos_cands : torch.Tensor
            The set of persona embeddings that co-occur with the dialog.
            
        neg_cands : torch.Tensor
            A set of persona embeddings negatively sampled from a list of candidates.

        Returns
        -------
        total_loss : torch.Tensor
            Loss on single sample over candidate comparisons.

        """
        num_pos, num_neg = pos_cands.size(0), neg_cands.size(0)
        total_loss = 0.
        count = 0
        for i in range(0, num_neg, num_pos):
            if pos_cands.size(0) == neg_cands[i:i+num_pos].size(0):
                total_loss = total_loss + self.ranking_loss(inp, pos_cands, neg_cands[i:i+num_pos])
                count +=1
        return total_loss
        
    def forward(self, states, positives, candidates):
        """
        Calculates the dialog history embeddings and outputs the triplet loss on batch. 
        
        Parameters
        ----------
        states : torch.Tensor
            Embeddings of input dialog trajectories. 
            
        positives : List of list strings
            List of person 2 persona facts corresponding to each conversation in the batch.    
            
        candidates : List of strings
            List of supporting set of persona facts to generate candidates from.
            
        Returns
        -------
        loss : float
            Triplet loss on batch.

        """
        # generate candidates using negative sampling
        pos_targets, neg_targets = self._generate_candidates(candidates, positives)
        # calculate ranking loss over candidates
        loss = sum([self._calculate_loss(states[i].unsqueeze(0), pos_targets[i], neg_targets[i])
                                for i in range(len(states))])
        return loss
    
    def score(self, states, positives, candidates):
        """
        Calculates precision and recall metrics based on embeddings of dialog trajectories.
        
        Parameters
        ----------
        states : torch.Tensor
            Embeddings of input dialog trajectories. 
            
        positives : List of list strings
            List of person 2 persona facts corresponding to each conversation in the batch.    
            
        candidates : List of strings
            List of supporting set of persona facts to generate candidates from.

        Returns
        -------
        prec1 : float
            # hits with the top guess / # samples in the batch. Measures quality of the top ranked guess.
            
        prec5 : float
            Average precsion@5 score, defined as # hits in top 5 guesses / 5. 
            
        rec5 : float
            Average recall@5 score, defined as # hits in top 5 guesses / # positives. Measures diversity of guesses.
            
        rec10 : float
            Average recall@5 score, defined as # hits in top 10 guesses / # positives. Measures diversity of guesses.

        Examples
        -------
        >>> from convogym.representers import StateEstimator
        >>> from convogym._decoder import model
        >>> dialog_history = ["hi how are you doing?",  # corresponds to person 2
                              "good, just got back from playing baseball.",
                              "i like it. i also like to play the piano."]
        >>> persona = ["i play the piano.", "i'm a little league all-star."]
        >>> state_estimator = StateEstimator(model)
        >>> ranking_loss = RankingLoss()
        >>> states = state_estimator(dialog_history) # outputs a (1 x 1024) tensor
        >>> prec1, prec5, rec5, rec10 = ranking_loss.score(states, [persona]) 
        >>> print("precision@5: %.2f, recall@5: %.2f" %(prec5, rec5))
        precision@5: 0.40, recall@5: 1.00
        
        """
        pos_targets, neg_targets = self._generate_candidates(candidates, positives)
        # calculate precision@k and recall@k metrics
        prec1 = [self._score_triplet(states[i].unsqueeze(0), pos_targets[i], neg_targets[i], at_k=1)
                 for i in range(len(states))]
        prec5 = [self._score_triplet(states[i].unsqueeze(0), pos_targets[i], neg_targets[i], at_k=5)
                 for i in range(len(states))]
        rec5 = [self._score_triplet(states[i].unsqueeze(0), pos_targets[i], 
                neg_targets[i], at_k=5, use_recall=True) for i in range(len(states))]
        rec10 = [self._score_triplet(states[i].unsqueeze(0), pos_targets[i], 
                neg_targets[i], at_k=10, use_recall=True) for i in range(len(states))]

        return prec1, prec5, rec5, rec10
    
        
    def _generate_candidates(self, candidates, positives):
        """
        Generates negative samples for triplet rankings.
        """
        negatives = [random.sample( set(candidates) - set(pos_set), self.k) for pos_set in positives]
        pos_samples = [self.state_estimator._get_persona_embeddings(p) for p in positives]
        neg_samples = [self.state_estimator._get_persona_embeddings(n) for n in negatives]
        return pos_samples, neg_samples
    
    def _score_triplet(self, x, pos,neg, at_k, use_recall=False):
        """
        Scores the ranking model based on the ranked list of generated candidates. 
        
        If use_recall then recall@k is used. Otherwise precision@k is used.
        """
        dist = nn.CosineSimilarity()
        scores = []
        for k,v in enumerate(pos):
            scores.append((k,1, dist(x,v.unsqueeze(0)).item()))
        for k,v in enumerate(neg):
            scores.append((k+len(pos), 0, dist(x,v.unsqueeze(0)).item()))
        scores = sorted(scores, key = lambda x: x[-1], reverse=True)
        picks = [s[1] for s in scores[0:at_k]]
        # precision or recall @ k
        score = sum(picks) / len(pos) if use_recall else sum(picks) / len(picks)
        return score

# interface with gym callback objects
class Reward:
    """
    A helper object for callbacks that outputs turn-level rewards based on state information at each turn.
    
    In the base Reward, a custom criterion function is 
    
    Parameters
    -------
    state_estimator : StateEstimator object or Callable
        A state estimator model that maps from dialog history text -> embedding vector to represent state information.
    
    criterion : nn.Module or Callable
        Maps from state vector to a scalar reward. The default is None.
    
        If want (state, action) -> reward, simply concatenate action to state vector as input.

    """
    def __init__(self, state_estimator, criterion):
        self.state_estimator = state_estimator
        self.criterion = criterion
    
    def calculate_reward(self, scb, mcb):
        """
        Obtains reward from given persona and state information and passes it to the state callback.
    
        Attributes
        -------
        scb.reward : float
            A scalar reward for current turn.
            
        scb.rewards: List of floats
            List of reward for each turn.
    
        """
        loss = self.criterion(scb.state.view(1,-1))
        scb.reward = - loss.item()
        scb.rewards.append(scb.reward)
        return scb, mcb
    
    def score_trajectory(self, scb, mcb):
        """
        Not implemented in base class.
        """
        raise NotImplementedError
        
class ManualReward(Reward):
    """
    Implements a wrapper on the Reward base object. 
    
    At each turn, the user is manually prompted to score the dialog up to the current turn. 
    (state, action, next_state, reward) samples collected this way can be used to learn reward function that approximates user scoring (i.e., a criterion). 
    """
    def __init__(self, state_estimator, *args, **kwargs):
        super().__init__(state_estimator=state_estimator, criterion=None)

    def calculate_reward(self, scb, mcb):
        """
        Obtains reward from given persona and state information and passes it to the state callback.
    
        Attributes
        -------
        scb.reward : float
            A scalar reward for current turn.
            
        scb.rewards: List of floats
            List of reward for each turn.
    
        """
        reward = None
        while reward is None:
            display_dialog_history(mcb.dialog_hx, self.state_estimator.tokenizer)
            try: 
                reward = float(input( "Enter a reward (float) for current dialog: " ))
            except:
                reward = None
        scb.reward = reward
        scb.rewards.append(scb.reward)
        return scb, mcb
    
    def score_trajectory(self, scb, mcb):
        """
        No additional metrics for this wrapper.
        """
        return scb, mcb

class RankingReward(Reward):
    """
    Implements a wrapper on the Reward base object. 
    
    Calculates various precision@k and recall@k metrics based on the RankingLoss criterion.
    
    Parameters
    -------
    state_estimator : StateEstimator object or Callable
        A state estimator model that maps from dialog history text -> embedding vector to represent state information.

    path_to_persona_facts : os.path or string, optional
        Path to list of persona facts. Used to generate candidates for ranking / scoring. 
        If none, a list of persona facts is constructed from the PersonaChat data set and saved locally.
        The default is none.
    """
    def __init__(self, state_estimator, path_to_persona_facts=None):
        super().__init__(state_estimator, None)
        self.criterion = RankingLoss(state_estimator)
        self.candidates = load_persona_facts(path_to_persona_facts)
        self.metrics = {'prec1s': [], 'prec5s': [],
                        'rec5s': [], 'rec10s': [], 'rewards': []}
    def calculate_reward(self, scb, mcb):
        """
        Obtains reward from given persona and state information and passes it to the state callback.
    
        Attributes
        -------
        scb.reward : float
            A scalar reward for current turn.
            
        scb.rewards: List of floats
            List of reward for each turn.
    
        """
        loss = self.criterion(states=scb.state.view(1,-1), 
                              positives=[scb.personas], 
                              candidates=self.candidates)
        scb.reward = - loss.item()
        scb.rewards.append(scb.reward)
        return scb, mcb
    
    def score_trajectory(self, scb, mcb):
        """
        Calculates precision@k and recall@k for current dialog trajectory.
    
        Parameters
        ----------            
        scb : StateCb object
            State callback object.
    
        Metrics
        -------
        prec1 : float
            # hits with the top guess / # samples in the batch. Measures quality of the top ranked guess.
            
        prec5 : float
            Average precsion@5 score, defined as # hits in top 5 guesses / 5. 
            
        rec5 : float
            Average recall@5 score, defined as # hits in top 5 guesses / # positives. Measures diversity of guesses.
            
        rec10 : float
            Average recall@5 score, defined as # hits in top 10 guesses / # positives. Measures diversity of guesses.
    
        """
        prec1, prec5, rec5, rec10 = self.criterion.score(states=scb.state.view(1,-1), 
                                                        positives=[scb.personas], 
                                                        candidates=self.candidates)
        # log data
        self.metrics['prec1s'].extend(prec1)
        self.metrics['prec5s'].extend(prec5)
        self.metrics['rec5s'].extend(rec5)
        self.metrics['rec10s'].extend(rec10)
        self.metrics['rewards'].append(scb.rewards)
        print("prec@1: %.2f | prec@5: %.2f | rec@5: %.2f | rec@10: %.2f" % (prec1[0], prec5[0], rec5[0], rec10[0]))
        print("avg p@1: %.2f | avg p@5: %.2f | avg r@5: %.2f | avg. r@10: %.2f" % (np.mean(self.metrics['prec1s']),
                                                                                   np.mean(self.metrics['prec5s']), 
                                                                                    np.mean(self.metrics['rec5s']), 
                                                                                    np.mean(self.metrics['rec10s'])))
        print()
        



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Aug 20 21:20:40 2021

@author: af1tang
"""
import os
import warnings
import random
import torch, torch.nn as nn
import numpy as np
from convogym.rewards import RankingLoss
from convogym.utils._device import to_var, to_device, device
from convogym.utils._reshape import chunker, flatten
from convogym.utils._visualization import plot_losses
from convogym._configs import opts

# embedding models
class BaseStateEstimator(nn.Module):
    """
    Base state estimator. 
    
    Uses a transformer model and tokenizer to map dialog tokens to a single hidden state.
    
    Parameters
    -------
    model : A Huggingface transformer decoder model (default = GPT2LMHeadModel).
        Currently supports Pytorch Huggingface (transformers) models only. 
        The default model is af1tang/personaGPT model card.
        
    tokenizer : Huggingface transformer tokenizer (default = GPT2Tokenizer).
        The corresponding tokenizer for the transformer model. 
        The default tokenizer is af1tang/personaGPT.
    
    """
    def __init__(self, model, tokenizer):
        super(BaseStateEstimator, self).__init__()
        self.model = model
        self.tokenizer = tokenizer
    
    def _get_hidden_state(self, dialog_history, reduction = "mean"):
        self.model.eval()   # fixes model to eval mode
        with torch.no_grad():
            hidden_states = self.model(to_var(flatten(dialog_history)).long(), 
                         output_hidden_states=True)[2][24].squeeze(0)
            if reduction == 'first':
                x = hidden_states[0]
            elif reduction == 'last':
                x = hidden_states[-1]
            else:
                x = hidden_states.mean(0)
        return x
    
    def forward(self, dialog_history, reduction = "mean"):
        """
        Converts a dialog history of a conversation into a single embedding vector in R^n.

        Parameters
        ----------
        dialog_hx : List of list of ints
            List of token ids, each corresponding to a dialog response.
            
        reduction : string, "mean", "first", "last" 
            Different ways to take a representative hidden state. 
            `mean` takes the average of the last hidden layer. 
            `first` takes the first token hidden state of the last hidden layer (e.g., BERT uses CLS token to summarize text).
            `last` takes the last token hidden state of the last hidden layer. 
            The default is "mean".

        Returns
        -------
        x : torch.Tensor
            Outputs a hidden state for the entire dialog history.

        """
        return self._get_hidden_state(dialog_history, reduction)
        
    def load_params(self, *args, **kwargs):
        """
        Not Implemented in the base class.
        """
        raise NotImplementedError

    def fit(self, *args, **kwargs):
        """
        Not Implemented in the base class.
        """
        raise NotImplementedError
        
    def evaluate(self, *args, **kwargs):
        """
        Not Implemented in the base class.
        """
        raise NotImplementedError
    
class StateEstimator(BaseStateEstimator):
    """
    Embedding neural network model that converts dialog history strings -> an embedding vector.
    
    The embedding of the dialog history is a "state representation".
    In Markov Decision Process terms, this is the state input to the policy
    The output shape is (1 x n) for a vector in R^n.
    
    Parameters
    -------
    model : A Huggingface transformer decoder model (default = GPT2LMHeadModel).
        Currently supports Pytorch Huggingface (transformers) models only. 
        The default model is af1tang/personaGPT model card.
        
    tokenizer : Huggingface transformer tokenizer (default = GPT2Tokenizer).
        The corresponding tokenizer for the transformer model. 
        The default tokenizer is af1tang/personaGPT.
            
    from_pretrained : string or os.path, optional
        Save and loading path to the embedder network state_dict (weights).
    
    inp_size : int, optional
        Size of the state space. The default is 1024.
    
    hidden_size : int, optional
        Number of hidden units per layer in the embedder model. The default is 1024.
        
    dropout : float, optional
        Dropout (0-1.0) between hidden layers for the embedder model. The default is 0.2.
        
    Attributes
    -------
    embedder : Callable, nn.Module or nn.Sequential
        Embedding neural network model for dialog history embeddings:
            - takes as input the averaged dialog history tokens 
            - outputs an embedding (feature vector) in R^n.
        
    Notes
    -----  
    The fit() and evaluate() methods train the identifier network on persona data (e.g., PersonaChat). 
    The score() method can potentially handle arbitruary dialog trajectories and personas 
    so long as the persona facts are found in self.p2v. 
    """
    def __init__(self, model, tokenizer, 
                 from_pretrained=None,
                 inp_size=1024, hidden_size=1024, dropout=.2):
        super().__init__(model=model, tokenizer=tokenizer)
        # network layers
        self.embedder = nn.Sequential( 
                nn.Linear(inp_size, hidden_size),
                nn.Tanh(),
                nn.Dropout(dropout),
                nn.Linear(hidden_size, hidden_size)
            ).to(device)
        
        self.checkpoint_path = from_pretrained
        if from_pretrained:
            self.load_params()
        else:
            warnings.warn("This state estimator has NOT been trained yet.")
        
    def forward(self, dialog_history, reduction = "mean"):
        """
        Converts a dialog history of a conversation into a single embedding vector in R^n.

        Parameters
        ----------
        dialog_hx : List of list of ints
            List of token ids, each corresponding to a dialog response.
            
        reduction : string, "mean", "first", "last" 
            Different ways to take a representative hidden state. 
            `mean` takes the average of the last hidden layer. 
            `first` takes the first token hidden state of the last hidden layer (e.g., BERT uses CLS token to summarize text).
            `last` takes the last token hidden state of the last hidden layer. 
            The default is "mean".
            
        Returns
        -------
        torch.Tensor
            Outputs an embedding vector for the entire dialog history. Applies embedding network on hidden state of the dialog history.

        """
        self.embedder.eval()
        return self.embedder(self._get_hidden_state(dialog_history, reduction))

    def _get_persona_embeddings(self, persona_facts, reduction="mean"):
        """
        Converts a set of persona facts from string -> embedding vectors in R^n.

        Parameters
        ----------
        persona_facts : List of strings
            List of persona facts in string form.

        reduction : string, "mean", "first", "last" 
            Different ways to take a representative hidden state. 
            `mean` takes the average of the last hidden layer. 
            `first` takes the first token hidden state of the last hidden layer (e.g., BERT uses CLS token to summarize text).
            `last` takes the last token hidden state of the last hidden layer. 
            The default is "mean".

        Returns
        -------
        vectors : torch.Tensor
            Outputs an embedding vector for each persona fact.

        """
        self.model.eval()   # fixes model to eval mode
        vectors = []
        with torch.no_grad():
            for line in persona_facts:     
                outp = self.model(**self.tokenizer(line, return_tensors='pt').to(device), output_hidden_states=True)
                x = outp[2][24].squeeze(0)
                if reduction=='first':
                    vectors.append(x[0])
                elif reduction=='last':
                    vectors.append(x[-1])
                else:
                    vectors.append(x.mean(0))
            vectors = torch.stack(vectors)
        return vectors
    
    def load_params(self):
        """
        Try to load the parameters of the identifier model using load_path. Raises an exception if fails.
        """
        try: 
            state_dict = torch.load(self.checkpoint_path)
            self.embedder.load_state_dict(state_dict)
        except Exception as e:
            warnings.warn(str(e) + '''
                          
                          This state estimator model is NOT trained yet.
                          
                          ''')

class RankingStateEstimator(StateEstimator):
    """
    A wrapper object over the base StateEstimator class that trains the base estimator using the RankingLoss object.
    
    Parameters
    -------
    model : A Huggingface transformer decoder model (default = GPT2LMHeadModel).
        Currently supports Pytorch Huggingface (transformers) models only. 
        The default model is af1tang/personaGPT model card.
        
    tokenizer : Huggingface transformer tokenizer (default = GPT2Tokenizer).
        The corresponding tokenizer for the transformer model. 
        The default tokenizer is af1tang/personaGPT.
            
    checkpoint_path : string or os.path, optional
        Save and loading path to the embedder network state_dict (weights).
    
    inp_size : int, optional
        Size of the state space. The default is 1024.
    
    hidden_size : int, optional
        Number of hidden units per layer in the embedder model. The default is 1024.
        
    dropout : float, optional
        Dropout (0-1.0) between hidden layers for the embedder model. The default is 0.2.
        
    k : int, optional
        Number of candidates to use for triplet loss. The default is 20.
        
    Attributes
    -------
    embedder : Callable, nn.Module or nn.Sequential
        Embedding neural network model for dialog history embeddings:
            - takes as input the averaged dialog history tokens 
            - outputs an embedding (feature vector) in R^n.
            
    criterion : RankingLoss
        Loss function used to train the state estimator network.
        
    """
    def __init__(self, model, tokenizer, 
                 from_pretrained=None,
                 inp_size=1024, hidden_size=1024, dropout=.2, k=20):
        super().__init__(model=model, tokenizer=tokenizer, 
                         from_pretrained=from_pretrained,
                         inp_size=inp_size, hidden_size=hidden_size,
                         dropout=dropout)
        self.criterion = RankingLoss(state_estimator=self, k=k)
        
    def fit(self, train_data, candidates,
                epochs=1, lr=1e-3, bs=32, 
                logging_steps=10, save_steps=50):
        """
        Trains the identifier model on a training set of (dialog token embeddings, persona facts) pairs.

        Parameters
        ----------
        train_data : List of tuples
            Batches of training samples of the form (dialog_history, persona 1, persona 2):
                - dialog_history: List of list of ints representing token ids for responses in conversation.
                - persona 1: list of strings corresponding to person 1 persona facts
                - persona 2: list of strings corresponding to person 2 persona facts.
                
        candidates : List of strings
            List of possible persona facts from which to do negative sampling.
            
        save_path : String or os.path
            Path to save identifier paramters to.
            
        epochs : int, optional
            Number of training epochs. The default is 1.
            
        lr : float, optional
            Learning rate. The default is 1e-3.
            
        bs : int, optional
            Batch size. The default is 32.
            
        logging_steps : int, optional
            Print training results every logging_steps. The default is 10.
            
        save_steps : int, optional
            Save model checkpoint every save_steps. The default is 50.

        Returns
        -------
        None.

        """
        print()
        print("Training the state estimator ... ")
        print()
        optimizer = torch.optim.Adam(self.embedder.parameters(), lr=lr)
        # training
        stats, iters, tr_loss, logging_loss = {}, 0, 0.0, 0.0
        self.embedder.train()
        self.embedder.zero_grad()
        iters = 0
        for epoch in range(epochs):
            random.shuffle(train_data)
            for minibatch in chunker(train_data, bs):
                # batching
                convos, p1, p2 = zip(*minibatch)
                # forward
                x1 = torch.stack([self.forward(convo[::2]) for convo in convos], dim=0)
                x2 = torch.stack([self.forward(convo[1::2]) for convo in convos], dim=0)
                loss1 = self.criterion(states=x1, positives=p1, candidates=candidates)
                loss2= self.criterion(states=x2, positives=p2, candidates=candidates)
                loss = loss1+loss2
                # backward
                loss.backward()
                tr_loss += loss.item()
                #if (step+1)% opts.gradient_accumulation_steps == 0:
                optimizer.step()
                self.embedder.zero_grad()
                iters +=1
                # reporting
                if iters % logging_steps ==0:
                    stats[iters] = {'loss': (tr_loss - logging_loss) / logging_steps}
                    logging_loss = tr_loss
                    print('Epoch: %d | Iter: %d | loss: %.3f ' %( 
                    epoch, iters, stats[iters]['loss']) )
                    
                if iters % save_steps==0:
                    print("Saving stuff ... ")
                    state_dict = self.embedder.state_dict()
                    torch.save(state_dict, self.checkpoint_path)
                    plot_losses(stats, title='loss' )
                    print("Done.")
                    
    def evaluate(self, test_data, candidates, bs=32):
        """
        Evaluates identifier model on test persona data.

        Parameters
        ----------
        test_data : List of tuples
            Batches of training samples of the form (dialog_history, persona 1, persona 2):
                - dialog_history: List of list of ints representing token ids for responses in conversation.
                - persona 1: list of strings corresponding to person 1 persona facts
                - persona 2: list of strings corresponding to person 2 persona facts.
                
        bs : int, optional
            Batch size. The default is 32.
        k : int, optional
            Number of candidates to use for triplet loss. The default is 20.

        Returns
        -------
        eval_stats : dict
            Dictionary of precision@k and recall@k stats.

        """
        # eval
        print("-"*50)
        print("Evaluating the state estimator ... ")
        eval_stats = {'prec@1':[], 'prec@5':[], 'rec@5':[], 'rec@10':[]}
        self.embedder.eval()
        for minibatch in chunker(test_data, bs):
            # batching
            convos, p1, p2 = zip(*minibatch)
            with torch.no_grad():
                # forward
                x1 = torch.stack([self.forward(convo[::2]) for convo in convos], dim=0)
                x2 = torch.stack([self.forward(convo[1::2]) for convo in convos], dim=0)
            for xx,yy in [(x1,p1), (x2,p2)]:
                prec1, prec5, rec5, rec10 = self.criterion.score(inp=xx, positive=yy, candidates=candidates)
                eval_stats['prec@1'].extend(prec1)
                eval_stats['prec@5'].extend(prec5)
                eval_stats['rec@5'].extend(rec5)
                eval_stats['rec@10'].extend(rec10)
        print("prec1: %.1f | prec5: %.1f | rec5: %.1f | rec10: %.1f" % ( 
                                100*np.mean(eval_stats['prec@1']), 100*np.mean(eval_stats['prec@5']), 
                                100*np.mean(eval_stats['rec@5']), 100*np.mean(eval_stats['rec@10'])) )
        print("-"*50)
        return eval_stats


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Aug 11 20:25:09 2021

@author: af1tang
"""
import os
import torch, random, pandas as pd
from convogym._configs import opts
from convogym.load_data import default_action_space as action_space
from convogym.prefixes import ( train_personas, test_personas,
                        get_custom_persona, get_random_persona, 
                        get_sequence_personas
                        )
from convogym.utils._visualization import display_dialog_history, to_tokens
from convogym.utils._reshape import flatten 

from convogym.learners import Learner
from convogym.agents import Agent
from convogym.callbacks import StateCb, MessageCb


## gym environments ##
class Gym:
    """
    The base Gym environment object used to train conversational agents. 
    
    A generative decoder (self.model) is used to decode turn-level responses. 
    
        - "user" (self.user) refers to the person 1 (initiator of convesation).
        - "agent" (self.agent) referes to person 2 (conversational partner).
        
    The self.user object may correspond to user inputs if self.interactive=True. 
    Otherwise, self.user is another Agent object (self.model parameterized by another set of persona facts). 

    Conversational histories and personality parameters are recorded for both the agent and the user.
    
    Parameters
    ----------
    model : A Huggingface transformer decoder model (default = GPT2LMHeadModel).
        Currently supports Pytorch Huggingface (transformers) models only. 
        The default model is af1tang/personaGPT.
        
    tokenizer : Huggingface transformer tokenizer (default = GPT2Tokenizer).
        The corresponding tokenizer for the transformer model. 
        The default tokenizer is af1tang/personaGPT.
        
    user : Agent object, optional
        If interactive=True, user is set to a function that takes user input strings as responses. 
        Otherwise, a new Agent object is initialized with a persona sampled from self.reset_persona_func.
        (default = None)
    
    interactive : bool, optional
        Whether to use human interaction. 
        (default = True)
        
    reset_persona_func : partial function, optional
        A sampling function that generates a set of persona facts from list of persona profiles.
        (default = get_custom_persona)
        
    length : int, optional
        Length of a given conversation. 
        (default = 8)
        
    top_k : int, optional
        Parameter to control number of candidates during top-k sampling at each decoding step. 
        (default = 10)
        
    top_p : float32, optional
        Float32 between [0.0 - 1.0] used for nucleus sampling. The 
        (default = 0.92)
        
    max_length : int, optional
        Maximum number of tokens allowed in conversation. 
        (default = 1024)

    Attributes
    -------
    data : dictionary
        A dictionary of dialog histories and persona facts for each conversation conducted.
            - hx: list of strings
                List of strings corresponding to each turn response.
            - p1: list of strings
                List of strings corresponding to the person facts of the first person.
            - p2: list of strings
                List of strings corresponding to the person facts of the second person.

    """
    def __init__(self, model, tokenizer, user=None, interactive = True,  
                 reset_persona_func=get_custom_persona, 
                 length=8, top_k=10, top_p = .92, max_length=1024 ):
        
        self.model = model
        self.tokenizer = tokenizer
        self.user = user
        self.interactive = interactive
        self.length, self.top_k, self.top_p, self.max_length = length, top_k, top_p, max_length
        self.reset_persona_func = reset_persona_func
        self.data = {'hx': [], 'p1': [], 'p2': []}
    
    def _reset_agents(self, persona_input, agent_personas, user_personas):
        """
        Reset self.agent and self.user Agent objects to handle conversations. 

        Parameters
        ----------
        persona_input : List of Strings
            A set of input personas to parameterize the agent (conversational partner). 
            If None, it is sampled from self.reset_persona_func.
            
        agent_personas : List of list of strings
            List of person profiles (each 3-5 strings of persona facts). 
            
            example: 
                - _personas.train_personas used during training self.agent
                - _personas.test_personas used during evaluation of self.agent
            
            (default = _personas.train_personas) 
            
        user_personas : List of list of strings
             List of person profiles (each 3-5 strings of persona facts).
            
            example:
                -  _personas.train_personas used during both training and testing of self.agent
            (default = _personas.train_personas) 
            
        Returns
        -------
        None.

        """
        if persona_input:
            self.agent = Agent(model=self.model, tokenizer=self.tokenizer, 
                               persona=persona_input, top_k=self.top_k,
                               top_p=self.top_p, max_length=self.max_length)
        else:
            persona = self.reset_persona_func(persona_list=user_personas)
            self.agent = Agent(model=self.model, tokenizer=self.tokenizer, 
                               persona=persona, reverse=False, top_k=self.top_k, 
                              top_p=self.top_p, max_length=self.max_length)
            
        if self.interactive:
            # human input
            self.user = self._interact
        else:
            # another persona model
            persona = self.reset_persona_func(persona_list=user_personas)
            self.user = Agent(model=self.model, tokenizer=self.tokenizer, 
                              persona=persona, reverse=True, top_k=self.top_k, 
                              top_p=self.top_p, max_length=self.max_length)
        
    def _interact(self, msg, act):
        """
        Get user input for response. Only used when self.interactive=True.

        Parameters
        ----------
        model : A Huggingface transformer decoder model (default = GPT2LMHeadModel).
            Currently supports Pytorch Huggingface (transformers) models only. 
            The default model is af1tang/personaGPT model card.
            
        msg : String
            Bot input message as a list of integers (tokenized).
            
        act : bool
            Whether to use action prefix instead of persona prefix.
            NOT used during base Gym. 

        Returns
        -------
        outp : List of integers
            Tokenized (string -> int) list of response tokens.

        """
        if msg:
            print("Bot: {}".format(self.tokenizer.decode(msg, skip_special_tokens=True)))
        outp = self.tokenizer.encode(input(">> User: ") + self.tokenizer.eos_token)
        return outp
    
    def _on_convo_begin(self, scb, mcb):
        """
        Callback updates for state (scb) and message (mcb) callbacks before conversation begins.

        Default behavior updates state persona tracker with the agent persona facts. 
        
        Parameters
        ----------
        scb : StateCb object
            State callback object that tracks state information of conversation.
            See callbacks.StateCb documentation for details.
            
        mcb : MessageCb object
            Message callback object that tracks current message information of conversation.
            See callbacks.MessageCb documentation for details.

        Returns
        -------
        scb : StateCb object
            Updated callback object that tracks state information of conversation.
            See callbacks.StateCb documentation for details.
            
        mcb : MessageCb object
            Updated callback object that tracks current message information of conversation.
            See callbacks.MessageCb documentation for details.

        """
        scb.personas = self.agent.p2
        return scb,mcb
        
    def _on_user_begin(self, scb, mcb):
        """
        Callback updates for state (scb) and message (mcb) callbacks before generating user response.

        Parameters
        ----------
        scb : StateCb object
            State callback object that tracks state information of conversation.
            See callbacks.StateCb documentation for details.
            
        mcb : MessageCb object
            Message callback object that tracks current message information of conversation.
            See callbacks.MessageCb documentation for details.

        Returns
        -------
        scb : StateCb object
            Updated callback object that tracks state information of conversation.
            See callbacks.StateCb documentation for details.
            
        mcb : MessageCb object
            Updated callback object that tracks current message information of conversation.
            See callbacks.MessageCb documentation for details.

        """
        return scb, mcb
    
    def _on_user_end(self, scb,mcb):
        """
        Callback updates for state (scb) and message (mcb) callbacks after generating user response.

        Updates dialog history (mcb.dialog_hx) with encoded user message. 
        
        Parameters
        ----------
        scb : StateCb object
            State callback object that tracks state information of conversation.
            See callbacks.StateCb documentation for details.
            
        mcb : MessageCb object
            Message callback object that tracks current message information of conversation.
            See callbacks.MessageCb documentation for details.

        Returns
        -------
        scb : StateCb object
            Updated callback object that tracks state information of conversation.
            See callbacks.StateCb documentation for details.
            
        mcb : MessageCb object
            Updated callback object that tracks current message information of conversation.
            See callbacks.MessageCb documentation for details.

        """
        mcb.dialog_hx.append(mcb.msg)
        return scb, mcb
    
    def _on_agent_end(self, scb, mcb):
        """
        Callback updates for state (scb) and message (mcb) callbacks after generating agent response.

        Default behavior:
            - updates dialog history with agent response
            - checks whether conversation ends (based on turn count and length)
        
        Parameters
        ----------
        scb : StateCb object
            State callback object that tracks state information of conversation.
            See callbacks.StateCb documentation for details.
            
        mcb : MessageCb object
            Message callback object that tracks current message information of conversation.
            See callbacks.MessageCb documentation for details.

        Returns
        -------
        scb : StateCb object
            Updated callback object that tracks state information of conversation.
            See callbacks.StateCb documentation for details.
            
        mcb : MessageCb object
            Updated callback object that tracks current message information of conversation.
            See callbacks.MessageCb documentation for details.

        """
        mcb.dialog_hx.append(mcb.msg)
        scb.turn +=1
        if scb.turn >= self.length:
            scb.done = True
        return scb, mcb
    
    def _on_convo_end(self, scb, mcb):
        """
        Callback updates for state (scb) and message (mcb) callbacks after conversation ends.

        Default behavior: 
            - prints dialog history 
            - updates self.data with dialog history, persona 1 and persona 2 facts.
            
        Parameters
        ----------
        scb : StateCb object
            State callback object that tracks state information of conversation.
            See callbacks.StateCb documentation for details.
            
        mcb : MessageCb object
            Message callback object that tracks current message information of conversation.
            See callbacks.MessageCb documentation for details.

        Returns
        -------
        scb : StateCb object
            Updated callback object that tracks state information of conversation.
            See callbacks.StateCb documentation for details.
            
        mcb : MessageCb object
            Updated callback object that tracks current message information of conversation.
            See callbacks.MessageCb documentation for details.

        """
        if not self.interactive:
            print('p1: ')
            print()
            for p in self.user.p1: print(p)
        print('-'*10)
        print('p2: ')
        print()
        for p in scb.personas: print(p)
        print('-'*10)
        print()
        display_dialog_history(mcb.dialog_hx, self.tokenizer)
        self.data['hx'].append(to_tokens( mcb.dialog_hx, self.tokenizer ))
        if self.interactive:
            self.data['p1'].append([])
        else:
            self.data['p1'].append(self.user.p1)
        self.data['p2'].append(scb.personas)
        del scb, mcb

    def _sim_convo(self, persona_input = None, 
                   agent_personas=None, user_personas=None): 
        """
        Generate a dialog trajectory consisting of 
            - dialog history
            - prefix information (persona or action prefixes) for each agent
            - state information associated w/ each turn (scb)

        Parameters
        ----------
        persona_input : List of strings, optional
            Used when custom persona is provided during self.interactive=True. The default is None.
            
        agent_personas : List of strings, optional
            Used when agent personas are sampled from a predfined list of personas. The default is None.
            
        user_personas : List of strings, optional
            Used when user personas are sampled from a predfined list of personas. for . The default is None.

        Returns
        -------
        None.

        """
        self._reset_agents(persona_input, agent_personas, user_personas)
        scb, mcb = StateCb(personas=agent_personas), MessageCb()
        scb, mcb = self._on_convo_begin(scb,mcb)
        # run for turns, updating rewards at each turn
        while not scb.done:
            # person 1 (user) 
            scb,mcb = self._on_user_begin(scb,mcb)             
            mcb.msg = self.user(mcb.msg, act=scb.act)
            scb,mcb = self._on_user_end(scb,mcb)

            # person 2 (agent)
            mcb.msg = self.agent(mcb.msg, act=scb.act)
            scb, mcb = self._on_agent_end(scb,mcb)
        self._on_convo_end(scb, mcb) 
    
    def sim_convos(self, num_convos=1, agent_personas=None, user_personas=None):
        """
        Simulates conversational episodes between user and agent dialog models using a given set of agent_personas and user_personas. 

        Parameters
        ----------
        num_convos : int, optional
            Number of conversations to simulate. The default is 1.
            
        agent_personas : List of strings, optional
            Used when agent personas are sampled from a predfined list of personas. The default is None.
            
        user_personas : List of strings, optional
            Used when user personas are sampled from a predfined list of personas. for . The default is None.


        Returns
        -------
        None.

        """
        print("Conducting conversations ...")
        print()
        for i in range(num_convos):
            self._sim_convo(None, agent_personas, user_personas)
 
    def save_data(self, save_path):
        """
        Save conversational episode data to file.

        Parameters
        ----------
        save_path : os.path or string
            Path to save self.data to. Saves as a .csv file. 

        Returns
        -------
        None.
        
        """
        df = pd.DataFrame(self.data)
        df.to_csv(save_path, index=False)
        
# active learning environments
class ActiveGym(Gym):
    """
    The active learning Gym environment is built from the base Gym environment. 
    
    In this setting, the user (human) chooses action prefixes rather than personas to control decoded responses.
        - Unlike persona prefixes ("<|p1|>", "<|p2|>") which delimits persona facts, 
            action prefix ("<|act|>") indicates the turn-level goals used to condition the language model for decoding.
        - Action space defines a set of "actions" (turn-level objective) that can be collected to train the model.
        - Active learning data can be saved for downstream training tasks. 
    
    Parameters
    ----------
    model : A Huggingface transformer decoder model (default = GPT2LMHeadModel).
        Currently supports Pytorch Huggingface (transformers) models only. 
        The default model is af1tang/personaGPT model card.
        
    tokenizer : Huggingface transformer tokenizer (default = GPT2Tokenizer).
        The corresponding tokenizer for the transformer model. 
        The default tokenizer is af1tang/personaGPT.
        
    training_data : torch.utils.data.Dataset or List of tuples , optional
        List of training batches to sample batches from. Used to re-fit model to prevent catastrophic forgetting.

    action_space : List of strings
        List of turn-level goals to train the decoder with using active learning.
        
    length : int, optional
        Length of a given conversation. 
        (default = 8)
        
    top_k : int, optional
        Parameter to control number of candidates during top-k sampling at each decoding step. 
        (default = 10)
        
    top_p : float32, optional
        Float32 between [0.0 - 1.0] used for nucleus sampling. The 
        (default = 0.92)
        
    max_length : int, optional
        Maximum number of tokens allowed in conversation. 
        (default = 1024)
        
    train_model : bool, optional
        Whether to fine-tune the model during active learning episodes. If False, the model is not fine-tuned between corrections.
        (default = True)
        
    lr : float, optional
        Learning rate for model parameters. The default is 5e-5.
        
    use_param_groups : bool, optional
        Whether to use different learn rates for special tokens, positional tokens and normal tokens.
        The default is True.
        
    schedule_func : torch.optim.lr_scheduler object, optional
        Learn rate scheduler function. The default is get_linear_schedule_with_warmup.
        
    gradient_accumulation_steps : int, optional
        Number of gradient accumulation steps. The default is 8.
        
    max_grad_norm : float, optional
        Max gradient norm size for gradient clipping. The default is 1.0.
        
    optim_func : torch.optim object, optional
        Optimizer used to update parameters. The default is AdamW.
        
    total_iters : int, optional
        Max number of training iters. The default is 20000.


    Attributes
    -------
    data : dict
        Active learning data in dictionary format: 
            - X: List of list of int
                Input tokens (encoded by tokenizer) of action prefix + dialog history.
            - y: List of list of int
                Human-provided response (ground truth).
            - dialog_hx: List of string
                List of responses at each turn.
            - actions: List of int
                List of actions (turn-level goals) represented by indices 
                
    learner : Learner object
        Training wrapper to update the decoder model (self.model) on active learning data.

    """
    def __init__(self, model, tokenizer, training_data=None, 
                 train_model=False, action_space=action_space,
                 length=8, top_k=10, top_p = .92, max_length=1024, 
                 # learner wrapper parameters
                 use_param_groups=False, lr=1e-3,
                 schedule_func=None, gradient_accumulation_steps=8, 
                 optim_func=None, total_iters=20000, max_grad_norm=1.0):
        
        super().__init__(model=model, tokenizer=tokenizer, user=None, interactive=True, 
                         reset_persona_func=get_sequence_personas,
                         length=length, top_k=top_k, top_p=top_p, max_length=max_length)
        self.train_model = train_model
        self.action_space = action_space
        if train_model:
            if not training_data:
                raise ValueError("training_set cannot be None while train_model = True.")
            self.learner = Learner(model = model, training_data = training_data, 
                                   lr = lr, 
                                   use_param_groups = use_param_groups,
                                   schedule_func=schedule_func,
                                   gradient_accumulation_steps=gradient_accumulation_steps,
                                   max_grad_norm=max_grad_norm,
                                   optim_func=optim_func,
                                   total_iters=total_iters)
            
        self.data = {'X': [], 'y': [], 'dialog_hx': [], 'actions': []}
    
    def _reset_agents(self, persona_input, agent_personas=None, user_personas=None):
        """
        In ActiveGym, the user is set to an Agent object without any persona facts. 
            - At each turn, the user provides an action (turn-level goal) to train the model with.
            - At the end of each conversation, a new persona profile (3-5 persona facts) is sampled from a list of training personas.
            
        """
        self.user = Agent(model=self.model, tokenizer=self.tokenizer, 
                          persona=[], top_k=self.top_k, top_p=self.top_p, max_length=self.max_length)
        self.agent = Agent(model=self.model, tokenizer=self.tokenizer, 
                            persona=persona_input, top_k=self.top_k,
                               top_p=self.top_p, max_length=self.max_length)
        
    def _on_convo_begin(self, scb, mcb):
        """
        At convo begin, set persona facts (ground truth) to agent persona facts. 

        Parameters
        ----------
        scb.personas : List of string
            Agent's persona facts set as ground truth. (User has not persona facts, uses action codes instead.)
            
        scb.record : bool
            Whether to record convo to self.data['dialog_hx'] and self.['actions'].
            (default = True)
            
        """
        scb.personas = self.agent.p2
        scb.record = True
        return scb,mcb
        
    def _on_user_begin(self, scb, mcb):
        """
        On user begin: 
            - User selects an action (turn-level goal) from action space. 
            - A candidate response is generated using the action code.
            - Updates message callback 'x' state (mcb.x) to the prefix tokens of the user agent.
                If first turn, '<|start|>' token is ommitted (used as first token in the response).
        """
        if mcb.msg is not None:
            mcb.dialog_hx.append(mcb.msg)
        action = None
        while action not in self.action_space:
            display_dialog_history(mcb.dialog_hx, self.tokenizer)
            print()
            print(" actions: ")
            for k,v in enumerate(self.action_space): print(k,v)
            try:
                int_act = int(input(" input [0-10]: " ))
                action = self.action_space[int_act]
            except:
                action = None            
        scb.action = action
        scb.actions.append(int_act)
        scb.act = True
        self.user.p1 = [scb.action]
        # cache current x
        x = self.tokenizer.encode(''.join(['<|act|> '] + self.user.p1 + ['<|sep|>'] + ['<|p1|>'] + [] + ['<|sep|>'] + ['<|start|>']))
        x += flatten(mcb.dialog_hx)
        x = torch.tensor([x])
        # set inp as input_ids for dataset
        if scb.turn == 0:
            x = x[:, :-1]
        mcb.x = x
        return scb, mcb
    
    def _on_user_end(self, scb,mcb):
        """
        On user end:
            - Check if decoded response is sensible and uses action (turn-level goals) correctly.
                - if satisfactory (y): continue conversation 
                - if unsatisfactory (n): user provides corrective response (ground-truth)
                
            - If user provides a corrected output, the conversation ends (scb.done=True) and the dialog history is not recorded (scb.record=False).
        
            - However, self.data is updated with the prefix tokens (mcb.x) and user response and added to the growing active learning dataset.
        
            - The last batch self.data['X'][-1] and self.data['y'][-1] is used as the active learning batch to update parameters.

        """
        mcb.dialog_hx.append(mcb.msg)
        # check if need revision
        print(); print('-'*50)
        display_dialog_history(self.user.dialog_history, self.tokenizer)
        print('-'*12, ' iter %d, turn %d '%(self.iter, scb.turn), '-'*12 )
        print("action: ", scb.action)
        decision = input(" continue? [y/n] ")
        # decision tree
        if decision == 'y':
            # augment even more turns to active data
            # self.data['X'].extend(mcb.x.tolist()); self.data['y'].append(mcb.msg)
            # continue conversation
            x = self.tokenizer.encode(''.join(['<|p2|>'] + scb.personas + ['<|sep|>'] + ['<|start|>']))
            x += flatten(mcb.dialog_hx)
            x = torch.tensor([x])
            mcb.x = x
        else:
            y = [[]]
            # get corrected user response
            while len(y[0]) < 2:
                y = self.tokenizer.encode(input("  >> user: ") + self.tokenizer.eos_token, return_tensors='pt')
            if scb.turn ==0:
                start_tok = self.tokenizer.encode('<|start|>', return_tensors='pt')
                y = torch.cat( (start_tok, y), -1)
            # extend active learning data and prepare active learning batch
            self.data['X'].extend( mcb.x.tolist() ); self.data['y'].extend( y.tolist())
            mcb.batch = (mcb.x,y)
            # retart convo
            scb.done, scb.record = True, False
        scb.act = False
        return scb, mcb
    
    def _on_agent_end(self, scb, mcb):
        """
        On agent end:
            - Check if decoded response is sensible and uses action (turn-level goals) correctly.
                - if satisfactory (y): continue conversation 
                - if unsatisfactory (n): user provides corrective response (ground-truth)
                
            - If user provides a corrected output, the conversation ends (scb.done=True) and the dialog history is not recorded (scb.record=False).
            
            - self.data is updated with the prefix tokens (mcb.x) and user response and added to the growing active learning dataset.
            
            - The last batch self.data['X'][-1] and self.data['y'][-1] is used as the active learning batch to update parameters.

        """
        if not scb.done:
            display_dialog_history(self.agent.dialog_history, self.tokenizer)
            print('-'* 12, ' iter %d, turn %d ' %(self.iter, scb.turn), '-' * 12)
            print(" personas: ")
            for p in scb.personas: print(p)
            decision = input( " continue? [y/n] " )
            if decision == 'y':
                scb.turn +=1
                if scb.turn >= self.length:
                    scb.done = True
            else:
                y = [[]]
                while len(y[0]) < 2:
                    y = self.tokenizer.encode(input("  >> user: ") + self.tokenizer.eos_token, return_tensors='pt')
                # extend active learning data and prepare active learning batch
                self.data['X'].extend( mcb.x.tolist() ); self.data['y'].extend( y.tolist())
                mcb.batch = (mcb.x,y)
                # retart convo
                scb.done, scb.record = True, False
        return scb, mcb
    
    def _on_convo_end(self, scb, mcb):
        """
        On conversation end: 
            - if scb.record (i.e., no user corrections), save dialog history and actions used
            - if user corrections, update model if training mode.
        """
        if scb.record:
            self.data['dialog_hx'].append(to_tokens(mcb.dialog_hx, self.tokenizer))
            self.data['actions'].append(scb.actions)
        elif self.train_model:
            self.learner.fit_on_active_batch(mcb.batch)
        del scb, mcb
        
    def sim_convos(self, num_epochs=3, num_convos=9999):
        """
        Generate dialog trajectories with 
            max number of episodes = min( size of persona set * epochs, num_convos). 
        """
        print("Conducting conversations ...")
        print()
        max_num_convos = min(len(train_personas) * num_epochs, num_convos)
        for self.iter, persona in enumerate(self.reset_persona_func(train_personas)):
            self._sim_convo(persona, None, None)
            if self.iter > max_num_convos:
                break
            
    def save_data(self, save_dir):
        """
        Save conversational episode data to file.

        Parameters
        ----------
        save_dir : os.path or string
            Directory to save self.data to. Saves as 2 .csv files.
                - active learning data (input context, target tokens)
                - imitation learning data (dialog history, actions)

        Returns
        -------
        None.
        
        """
        al_data = self.data['X'], self.data['y']
        il_data = self.data['dialog_hx'], self.data['actions']
        if al_data:
            al_df = pd.DataFrame(al_data, columns=['X', 'y'])
            al_df.to_csv(os.path.join(save_dir, 'active_learning_data.csv'), index=False)
        if il_data:
            il_df = pd.DataFrame(il_data, columns=['dialog_hx', 'actions'])
            il_df.to_csv(os.path.join(save_dir, 'imitation_learning_data.csv'), index=False)                    

# RL environments
class RLGym(Gym):
    """
    The reinforcement learning Gym environment is built from the base Gym environment. 
    
    In this setting, the user is a policy that learns to output a distribution over an action space (previously defined turn-level goals).
    The policy is trained to direct conversations toward dialog-level goals, which is represented as a reward object (self.reward_obj). 
    Like the active learning setting, the turn-level goals are incorporated as part of the input (i.e., as context) to the decoder model. 
    Unlike the action learning, the model is fixed, and the policy is trained to optimize over turn-level goals as actions rather than token-level outputs by the decoder model. 
    
    Parameters
    ----------
    model : A Huggingface transformer decoder model (default = GPT2LMHeadModel).
        Currently supports Pytorch Huggingface (transformers) models only. 
        The default model is af1tang/personaGPT model card.
        
    tokenizer : Huggingface transformer tokenizer (default = GPT2Tokenizer).
        The corresponding tokenizer for the transformer model. 
        The default tokenizer is af1tang/personaGPT.
        
    policy : Policy object
        The policy outputs a turn-level goal (e.g., "talk about work") to be incorporated into the user object (self.user). 
        See convogym.policies for details.
            
    env : Env object
        An environment object to estimate state information from dialog history.
        See convogym.environments for details.
        
    reward_obj : Reward object
        A reward object to track reward function outputs at each dialog turn.
        See convogym.rewards for details.
        
    max_buffer_size : int, optional
        Maximum replay buffer size to use for training the policy. The default is 1000.
        
    length : int, optional
        Length of a given conversation. 
        (default = 8)
        
    top_k : int, optional
        Parameter to control number of candidates during top-k sampling at each decoding step. 
        (default = 10)
    
    top_p : float32, optional
        Float32 between [0.0 - 1.0] used for nucleus sampling. The 
        (default = 0.92)
    
    max_length : int, optional
        Maximum number of tokens allowed in conversation. 
        (default = 1024)

    Attributes
    -------
    data : dictionary
        A dictionary of dialog histories and persona facts for each conversation conducted.
            - dialog_hx: list of strings
                List of strings corresponding to each turn response.
            - actions: list of int
                List of integers corresponding to index of the sampled action at each turn.
            - rewards: list of floats
                List of floats corresponding to the reward received at each turn.
            - personas: list of strings
                List of strings corresponding to the person facts of the agent (second) person.
    
    memory_buffer : List of tuple
        Current dataset of batch tuples of the form 
            (states, actions, next_states, next_acts, rewards). 
            
        (state, action) -> (next_state, next_action), reward is observed at each transition as a part of the Markov Decision Process. 
            
    """
    def __init__(self, model, tokenizer, policy, env, reward_obj, 
                 max_buffer_size=1000,
                 length=8, top_k=10, top_p = .92, max_length=1024):        
        super().__init__(model=model, tokenizer=tokenizer, user=None, interactive=False, 
                         reset_persona_func=get_sequence_personas,
                         length=length, top_k=top_k, top_p=top_p, max_length=max_length)
        # REQUIRE: policy object
        self.policy = policy
        self.R = reward_obj
        self.Env = env
        self.data = {'dialog_hx':[], 'actions': [], 'personas': [], 'rewards': []}
        self.memory_buffer, self.max_buffer_size = [], max_buffer_size
    
    def _reset_agents(self, persona_input, agent_personas=None, user_personas=None):
        """
        Resets Agent objects for user (no input personas, uses action prefixes) and agent (randomly sampled persona, uses persona prefixes).
        """
        self.user = Agent(model=self.model, tokenizer=self.tokenizer, 
                          persona=[], top_k=self.top_k, top_p=self.top_p, max_length=self.max_length)
        self.agent = Agent(model=self.model, tokenizer=self.tokenizer, 
                           persona=persona_input, top_k=self.top_k,
                               top_p=self.top_p, max_length=self.max_length)
    
    def _on_convo_begin(self, scb, mcb):
        """
        On conversation begin:
            Initializes the state for tracking using a StateCb callback (scb). 
            The default state objects are torch.Tensor feature vectors. 
        """
        scb = self.Env.reset(scb)
        scb.personas = self.agent.p2
        return scb, mcb
        
    def _on_user_begin(self, scb, mcb):
        """
        On user begin:
            - Obtains a state representation from current dialog history.
            - Policy maps state -> action (sampled from action space, no gradient).

        Parameters
        ----------
        int_act : int
            Sampled output (no grad) from policy output. Corresponds the index of an action in action space.
            
        action : string
            Sampled action.

        scb.act : bool
            Used in 
                - self.user._reset_inp(act=scb.act) to prepare input format to include the <|act|> token.
        """
        # sample action from policy
        inp = self.Env.get_policy_inp(scb)
        int_act = self.policy.act(inp, scb.turn)
        # update action
        scb.action = self.policy.action_space[int_act]
        scb.actions.append(int_act)
        scb.act = True
        self.user.p1 = [scb.action]
        return scb, mcb
    
    def _on_user_end(self, scb,mcb):
        """
        Updates dialog history, switches scb.act = False.
        
        Parameters
        -------
        scb.act : bool
            Used in self.agent._reset_inp(act=scb.act) to prepare input format to use persona prefixes (e.g., <|p2|>).
        """
        mcb.dialog_hx.append(mcb.msg)
        scb.act = False
        return scb, mcb
    
    def _on_agent_end(self, scb, mcb):
        """
        On agent end:
            - Track dialog history and update turn counts.
            - Get state, action -> next_state transition from environment. 
            - Get reward for dialog trajectory up to current turn.
        """
        mcb.dialog_hx.append(mcb.msg)
        # update state based on dialog history
        scb, mcb = self.Env.get_curr_state(scb, mcb)
        # calculate reward 
        scb, mcb = self.R.calculate_reward(scb, mcb)
        # transition to next state
        scb, mcb = self.Env.get_next_state(scb, mcb)
        # next turn
        scb.turn +=1
        if scb.turn >= self.length:
            scb.done = True
        return scb, mcb
    
    def _update_memory(self,batch):
        """
        Update memory buffer with current episode of batch tuples. If max buffer size exceeded, delete some episodes.
        """
        self.memory_buffer.extend([list(tuples) for tuples in batch])
        # delete from front of memory batch if full
        if len(self.memory_buffer) > self.max_buffer_size:
            self.memory_buffer = self.memory_buffer[len(batch):]
    
    def _on_convo_end(self, scb, mcb):
        """
        On conversation end:
            - Display dialog history, actions and personas.
            - Update self.data with trajectory data. 
            - Update memory buffer.
        """
        # display
        print('actions: ')
        print()
        for p in scb.actions: print(self.policy.action_space[p])
        print('-'*10)
        print('p2: ')
        print()
        for p in scb.personas: print(p)
        print('-'*10)
        print()
        display_dialog_history(mcb.dialog_hx, self.tokenizer)
        # evaluate dialog and log
        self.R.score_trajectory(scb, mcb)
        self.data['dialog_hx'].append(to_tokens(mcb.dialog_hx, self.tokenizer))
        self.data['actions'].append(scb.actions)
        self.data['personas'].append(self.agent.p2)
        self.data['rewards'].append(scb.rewards)
        # update memory buffer
        next_states = scb.states[1:]; states = scb.states[:-1]
        next_acts = scb.actions[1:] + [0]
        dones =  [False]*(len(scb.rewards)-1) + [True]
        batch = list(zip(states, scb.actions, next_states, next_acts, 
                         scb.rewards, dones))
        self._update_memory(batch)
        del scb, mcb
    
    def sim_convos(self, num_epochs=1, num_convos=9999, training=True):
        """
        Generate dialog trajectories. 
            - If training mode, update policy parameters at end of episode using gradient descent over memory buffer samples. 
            - If test mode, evaluate policy on test set personas, no gradient updates for policy.

        Parameters
        ----------
        num_convos : int, optional
            Max number of conversations to generate. The default is 9999.
            
        training : bool, optional
            Whether to update the policy after each generated trajectory. If True, train personas are used to parameterize agent. The default is True.
        """
        print("Conducting conversations ...")
        print()
        if training:
            max_num_convos = min(len(train_personas) * num_epochs, num_convos)
            for epoch in range(num_epochs):
                for self.iter, persona in enumerate(self.reset_persona_func(train_personas)):
                    print("="*20, "epoch %d, iter %d, training" %(epoch+1, self.iter), "="*20)
                    self._sim_convo(persona, None, None)
                    # policy update
                    print("Offline batch updates ... ")
                    self.policy.update_policy(self.memory_buffer)
                    print()
                    if self.iter > max_num_convos:
                        break
        else:
            max_num_convos = min(len(test_personas), num_convos)
            for self.iter, persona in enumerate(self.reset_persona_func(test_personas)):
                print("="*20, "iter %d, testing" %self.iter, "="*20)
                self._sim_convo(persona, None, None)
                if self.iter > max_num_convos:
                    break
        


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Aug 14 16:19:04 2021

@author: af1tang
"""
import os
import warnings
import random
import pandas as pd
from convogym._configs import opts

# default special tokens
eos_token = '<|endoftext|>'

# load train and test personas
tr_df = pd.read_csv(os.path.join(opts.data_path, 'train_personas.csv'))
te_df = pd.read_csv(os.path.join(opts.data_path, 'test_personas.csv'))

train_personas = []
for i, row in tr_df.iterrows():
    train_personas.append([fact + eos_token for fact in row.dropna().tolist()])

test_personas = []
for i, row in te_df.iterrows():
    test_personas.append([fact + eos_token for fact in row.dropna().tolist()])

    
# persona sampling functions 
def load_persona_facts(save_path=None):
    """
    Load a set of unique persona facts from save file.

    Parameters
    ----------
    save_path : os.path or string, optional
        Loads .csv format file with "Facts" as the column name for persona facts. 
        If none provided, constructs a list of unique persona facts from the PersonaChat dataset. The default is None.

    Returns
    -------
    persona_facts : List of strings
        List of unique persona facts from save file.

    """
    if save_path:
        try:
            df_facts = pd.read_csv(save_path)
            persona_facts = df_facts['Facts'].tolist()
        except: 
            try: 
                import pickle
                with open(save_path, 'rb') as f:
                    persona_facts = pickle.load(f)
            except Exception as e:
                warnings.warn( e )
                print(""" 
                      Unable to load person facts from file. 
                      
                      If file is .csv, make sure the persona facts column is labeled "Facts".
                      Otherwise try serializing the list of persona facts and saving it as a pickle object.
                      """)
    else:
        try:
            df_facts = pd.read_csv(os.path.join(opts.example_path, 'persona_facts.csv'))
            persona_facts = df_facts['Facts'].tolist()
        
        except Exception as e:
            warnings.warn(str(e))
            from load_data import prepare_state_estim_dataset
            from decoders import model, tokenizer
            print()
            print("Building unique persona facts from personachat dataset.")
            persona_facts = prepare_state_estim_dataset(model, tokenizer,
                                                       os.path.join(opts.data_path, 'personachat.csv'))
    return persona_facts

def get_custom_persona(n=3, save_path=None, *args, **kwargs):
    """
    Get a set of persona facts from user inputs.

    Parameters
    ----------
    n : TYPE, optional
        Number of persona facts to input. The default is 3.
        
    save_path : os.path or string, optional
        Path to save custom persona set to. If none, personas will not be saved. The default is None.

    Returns
    -------
    personas : List of strings.
        A single set of persona facts (3-5) from a list of personas.

    """
    personas = []
    for i in range(n):
        response = ""
        while len(response) <1:
            response = input(">> Fact %d: "%(i+1))+ eos_token
        personas.append(response)
    if save_path:
        try:
            df = pd.DataFrame(personas, columns=['Facts'])
            df.to_csv(save_path, index=False)
        except Exception as e:
            warnings.warn(e)
            print("Could not save new persona list to save file.")
    return personas

def get_random_persona(persona_list):
    """
    Samples a random persona profile (3-5 facts) from a list of input personas. 
    (Default = _personas.train_personas)

    Parameters
    ----------
    persona_list : List of list of strings.
        List of persona profiles from which to sample from.

    Returns
    -------
    List of strings.
        A single set of persona facts (3-5) from a list of personas.

    """
    if not persona_list:
        return random.sample(train_personas, 1)[0]
    else:
        return random.sample(persona_list, 1)[0]
        

def get_sequence_personas(persona_list):
    """
    Yields a set of persona facts (3-5) from a list of personas.
    (Default = _personas.train_personas)

    Parameters
    ----------
    persona_list : List of list of strings.
        List of persona profiles from which to sample from.

    Yields
    ------
    List of strings.
        A single set of persona facts (3-5) from a list of personas.

    Example 
    -----
    >>> from convogym._personas import train_personas
    >>> next(get_sequence_personas())
    
    ['i design video games for a living.<|endoftext|>',
     'i hate broccoli.<|endoftext|>',
     'i am afraid of the dark.<|endoftext|>',
     'my mom is my best friend.<|endoftext|>']
    """
    random.shuffle(persona_list)
    p_iter = iter(persona_list)
    while p_iter:
        try:
            yield next(p_iter)
        except:
            return


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Sep  9 13:32:22 2021

@author: af1tang
"""
import torch
from convogym.utils._device import device, to_var

def fit_on_batch(model, batch, gradient_accumulation_steps=8):
    """
    Calculate loss on batch, update gradient and backpropagate.

    Parameters
    ----------
    model : A Huggingface transformer decoder model (default = GPT2LMHeadModel).
        Currently supports Pytorch Huggingface (transformers) models only. 
        The default model is af1tang/personaGPT model card.
        
    batch : tuple of torch.LongTensors
        Format of a batch tuple is (xx,yy): 
            - xx: List of int
                Input tokens (encoded by tokenizer) of action prefix + dialog history.
            - yy: List of int
                Encoded human-provided response (ground truth).
        
    gradient_accumulation_steps : int, optional
        Number of gradient accumulation steps. The default is 8.

    Returns
    -------
    loss : float
        Loss on the current batch.

    """
    xx,yy = batch
    try:
        xx, yy = torch.stack(xx, -1).to(device), torch.stack(yy, -1).to(device)
    except:
        xx, yy = to_var(xx).long(), to_var(yy).long()
    ## forward on new data batch
    _outp = model(xx)
    past = _outp.past_key_values
    outp = model(yy, past_key_values=past, labels=yy)
    
    # backward
    loss = outp[0]; del outp
    if gradient_accumulation_steps > 1:
        loss = loss / gradient_accumulation_steps
    loss.backward()
    return loss

def _process_dqn_batch(batch, policy, swa_policy, gamma):
    """
    Processes a batch of (state, context, action, next_state, next_action) -> (inputs, actions, Q-targets).

    Parameters
    ----------
    batch : tuple of torch.Tensor objects
        (state, context, action, next_state, next_action), each a torch.Tensor object
        
    policy : Callable OR nn.Module
        The policy network that maps from (state, context) -> action.
        
    swa_policy : Callable or nn.Module
        The target network used to calculate the Q-target.
        
    gamma : float (0,1.0)
        Discount factor when calculating Q-values.

    Returns
    -------
    xx : torch.Tensor
        Policy input.
        
    y : torch.LongTensor
        Index of the action sampled from action space. The Q-value corresponding to the action index is used to calculate the gradient for backpropagation.
        
    q_targets : torch.Tensor
        Scalar Q-values are calculated by the target network across each sample in the batch. These are treated as labels (no gradients through the Q-targets) for temporal difference learning.

    """
    try:
        x, y, x_next, y_next, r, dones = batch
        y_next = to_var(y_next)

    except:
        x, y, x_next, y_next, r, dones = batch
    xx = to_var(torch.stack(x, dim=-1).type(torch.FloatTensor))# torch.stack(c, dim=-1).type(torch.cuda.FloatTensor)
    xx_next = to_var(torch.stack(x_next, dim=-1).type(torch.FloatTensor)) #torch.stack(c_next, dim=-1).type(torch.cuda.FloatTensor)
    y, r, dones = to_var(y), to_var(r.type(torch.FloatTensor)), to_var(dones.long())
    #xx = torch.cat((x,c), dim=-1)
    #xx_next = torch.cat((x_next, c_next), dim=-1)
    # calculate q-values
    with torch.no_grad():
        # use target network to predict q-targets
        q_values = policy(xx_next)
        idx = q_values.max(1)[1]
        q_values = swa_policy(xx_next)
        q_targets = r + (1-dones) * gamma * q_values[torch.arange(len(idx)), idx]
    return xx, y, q_targets



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Aug 19 14:21:16 2021

@author: af1tang
"""
from itertools import groupby

def filter_turn_indices(x, eos_token_id):
    filtered = [[t[1] for t in list(g)] for k,g in groupby(list(enumerate(x)), lambda x: x[1]==eos_token_id) if not k]
    return filtered


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Aug 19 14:20:35 2021

@author: af1tang
"""
import os
import matplotlib.pyplot as plt
from convogym._configs import opts, create_dir

def display_dialog_history(dialog_hx, tokenizer):
    for j, line in enumerate(dialog_hx):
        msg = tokenizer.decode(line)
        if j %2 == 0:
            print(">> User: "+ msg)
        else:
            print("Bot: "+msg)
            print()

def to_tokens(dialog_history, tokenizer):
    return [tokenizer.decode(line) for line in dialog_history]

### plotting ###    
def plot_losses(stats, title='loss'):
    create_dir(opts.plot_path)
    x = list(sorted(stats.keys()))
    loss = [stats[i][title] for i in x]
    plt.plot(x, loss, label= title)
    plt.legend()
    plt.title("%s" %title)
    plt.tight_layout()
    plt.savefig(os.path.join(opts.plot_path,'%s.png'%title))
    plt.close()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Aug 19 15:07:07 2021

@author: af1tang
"""
import os
import warnings
import random, math
import numpy as np
import torch, torch.nn as nn, torch.nn.functional as F
from torch.utils.data import DataLoader

from convogym.load_data import default_action_space as action_space
from convogym._configs import opts
from convogym.utils._optim_helpers import _process_dqn_batch
from convogym.utils._device import to_device
from convogym.utils._swa_utils import AveragedModel
from convogym.utils._visualization import plot_losses


from torch.optim import AdamW
from transformers import get_linear_schedule_with_warmup

default_policy = nn.Sequential(nn.Linear(1025, 512),
                        nn.Tanh(), nn.Dropout(.1),
                        nn.Linear(512, 512),
                        nn.Tanh(), nn.Dropout(.1),
                        nn.Linear(512, len(action_space)))

class Policy(nn.Module):
    """
    The Policy object is a base dialog policy that outputs a distribution over an action space (turn-level goals) given state information about a dialog trajectory.
    
    The base Policy object is not trainable. A wrapper object should be created with a new self.update_policy method to specify the training loop.
    
    Parameters
    ----------
    policy : torch.nn.Module, optional
        Pytorch network that maps from state space representation of dialog history to a distribution over action space. 
        The default is default_policy.
        
    lr : float, optional
        Learning rate for the policy network. The default is 1e-4.
        
    action_space : List of string, optional
        List of strings of action space. The default is action_space.
        
    from_pretrained : os.path or string, optional
        A checkpoint path to save or load the parameters of the policy.
        
    EPS_START : float, optional
        Float (0, 1.0) for the initial epsilon parameter in epsilon-greedy sampling. The default is .5.
        
    EPS_END : float, optional
        Float (0, 1.0) for the final epsilon parameter in epsilon-greedy sampling. The default is .5.
        
    EPS_DECAY : int, optional
        Number of steps between EPS_START and EPS_END. The default is 2048.

    Attributes
    -------
    huber_loss : nn.Module 
        Huber loss function used to calculate the Bellman Error during Q-learning. Not used unless Q-learning is the training method.
        
    action_space : List of strings
        List of possible actions that the policy can output. These correspond to turn-level goals that get incorpoated as prefix tokens in the decoder model for response generation.
    
    Examples
    -------
    >>> from convogym.policies import Policy # default policy
    >>> from convogym.utils._device import to_var
    >>> import torch, torch.nn as nn
    >>> action_space = [ 
        'talk about work.', 
        'ask about hobbies.',
        'talk about movies.'
        ]
    >>> policy_net = nn.Sequential(nn.Linear(100, len(action_space))
                                )
    >>> policy = Policy( policy = policy_net,
                        action_space = action_space) 
    >>> action = policy.act( inp=to_var(torch.rand((1,100))), turn=1 ) 
    >>> print("[%d] %s" %(action, action_space[action]))
    [1] ask about hobbies.        
    >>> inp = tokenizer.encode(''.join(['<|act|>'] + [action_space[action]] + ['<|sep|>'] + ['<|p1|>'] + ['<|sep|>'] + ['<|start|>']) )
    >>> print(inp)        
    [50262, 2093, 546, 45578, 13, 50257, 50260, 50257, 50259]        
    >>> tokenizer.decode(model.generate(to_var(inp).long().view(1,-1)).tolist()[0][len(inp):] )        
    'hello do you have any hobbies?<|endoftext|>'        

    """    
    def __init__(self, policy=default_policy,
                 from_pretrained=None,
                 lr=1e-4, action_space=action_space,
                EPS_START=.5, EPS_END=.05, EPS_DECAY=int(2e4)):
        super(Policy, self).__init__()
        self.policy = to_device(policy)
        # loss func
        self.huber_loss = nn.SmoothL1Loss()
        # epsilon sampling params
        self.EPS_START, self.EPS_END, self.EPS_DECAY = EPS_START, EPS_END, EPS_DECAY
        self.reset_stats()
        self.action_space = action_space
        # load params, warn if not trained
        self.checkpoint_path = from_pretrained
        if from_pretrained:
            self.load_params()
        else:
            warnings.warn("This policy has NOT been trained yet.")
        
    def reset_stats(self):
        """
        Reset global stats. 
        
        Attributes
        -------
        global_step : int
            Total number of actions (i.e., turns) taken in the environment.
        
        global_epoch : int 
            Total number of iterations of the training data.
            
        stats : dict
            Dictionary of losses at each batch update. Loss can be: 
                - reward in policy gradient
                - Belmman error in Q-learning
                - cross entropy in imitation learning.

        """
        # eps greedy params
        self.global_step = 0
        # logging
        self.stats, self.global_epoch = {'losses': {}}, 0
    
    def act(self, inp, turn):
        """
        Outputs a distribution over self.action_space and samples an action from the distribution.

        Parameters
        ----------
        inp : torch.tensor
            State representation of current dialog trajectory. The default representation is a (1, d) shaped torch.tensor feature vector in R^d.
            
        turn : int
            Current turn in conversation. In the initial turn, the action is sampled from the output distribution; otherwise the argmax action is chosen.

        Returns
        -------
        action : int
            Integer index of the action chosen from given action space.

        """
        self.policy.eval()
        with torch.no_grad():
            logits = self.policy(inp)
            p = F.softmax(logits,-1)
            # sampling w/ eps-greedy
            eps = random.random()
            threshold = self.EPS_END + (self.EPS_START - self.EPS_END) * math.exp(-1. * self.global_step / self.EPS_DECAY)
            self.global_step +=1
            if eps > threshold:
                if turn == 0:
                    action = torch.multinomial(p, num_samples=1).item()
                else:
                    action = p.max(1)[1].item()
            else:
                action = random.randrange(len(self.action_space))
        return action
    
    def update_policy(self, *args, **kwargs):
        """
        Not implemented
        """
        warnings.warn("update_policy() is called but not implemented.")

    def load_params(self):
        """
        Try to load the parameters of the identifier model using load_path. Raises an exception if fails.
        """
        try: 
            state_dict = torch.load(self.checkpoint_path)
            self.policy.load_state_dict(state_dict)
        except Exception as e:
            warnings.warn(str(e) + '''
                          
                          This policy model is NOT trained yet.
                          ''')
                
    def save(self):
        """
        Saves model parameters and config to save_path.
        """
        torch.save(self.policy.state_dict(), self.checkpoint_path)


class DQNPolicy(Policy):
    """
    A wrapper policy object that inherits from the base Policy object.
    
    DQNPolicy object specifies a training loop for the policy network using the self.update_policy() method. 
    
    New Attributes
    -------
    swa_policy : Callable or torch.nn.Module
        The target network used to calculate the Q-target.
        
    optimizer : torch.optim object
        Optimizer used to implement backpropagation on policy parameters.
        
    Scheduler: torch.optim.lr_scheduler object
        Scheduler used to change learning rate after each gradient step.
    
    """
    def __init__(self, policy=default_policy,
                 from_pretrained=None,
                 lr=1e-4, action_space = action_space, gamma=.95,
                EPS_START=.5, EPS_END=.05, EPS_DECAY=int(2e4), t_total=int(2e4)):
        super().__init__(policy=policy, from_pretrained=from_pretrained,
                         lr=lr, action_space=action_space, 
                         EPS_START=EPS_START, EPS_END=EPS_END, EPS_DECAY=EPS_DECAY)
        # optim and sched
        self.optimizer = AdamW(self.policy.parameters(), lr=lr)
        self.scheduler = get_linear_schedule_with_warmup(self.optimizer, num_warmup_steps=0, 
                                                num_training_steps=t_total)
        self.gamma = gamma
        # swa tools
        ema_avg = lambda averaged_model_parameter, model_parameter, num_averaged:\
                0.99 * averaged_model_parameter + 0.01 * model_parameter
        self.swa_policy = AveragedModel(self.policy, avg_fn = ema_avg)
        self.swa_policy.eval()
    
    def update_policy(self, memory_buffer, num_train_epochs=1):
        self.policy.train()
        for epoch in range(num_train_epochs):
            print()
            dataloader = DataLoader(memory_buffer, batch_size=64, shuffle=True)
            epoch_loss = []
            for i, batch in enumerate(dataloader):#dataloader_old):
                # get batch on buffer trajectories 
                xx, y, q_targets = _process_dqn_batch(batch, self.policy, self.swa_policy, self.gamma)
                # forward
                yhat = self.policy(xx)   # logits
                loss = self.huber_loss(yhat[torch.arange(len(y)),y], q_targets)      
                # backward
                self.policy.zero_grad()
                loss.backward()
                self.optimizer.step()
                # take swa (polyak avg)
                self.swa_policy.update_parameters(self.policy)
                # tracking
                epoch_loss.append(loss.item())
            print("epoch: %d | loss: %.2f | lr: %s" %(epoch, np.mean(epoch_loss), str(self.scheduler.get_last_lr()[-1])))
            # on epoch end
            self.scheduler.step()
            self.stats['losses'][self.global_epoch] = {'dqn_loss': np.mean(epoch_loss)}
            plot_losses(self.stats['losses'], title='dqn_loss' )
            self.global_epoch +=1


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Sep  8 19:06:05 2021

@author: af1tang
"""

from setuptools import setup, find_packages

setup( 
    name = "convogym",
    version = "0.1.2",
    url = "https://github.com/af1tang/convogym",
    author = "Fengyi (Andy) Tang",
    author_email = "af1tang2@gmail.com",
    description = """
    A gym environment to train conversational agents for custom tasks through 
    active learning and self-play.
    """,
    long_description = open('README.rst').read(),
    packages = ['convogym', 'convogym.utils', 'convogym.test'],
    package_data = {'convogym': ['data/personachat.csv','data/train_personas.csv', 'data/test_personas.csv']},
    install_require = [
        "pytorch==1.4.0",
        "numpy>=1.20.2",
        "scipy>=1.6.2",
        "pandas>=1.2.4",
        "transformers==4.10.0",
        "dotenv",
        "tqdm",
    ],
    classifiers = [
        "Programming Language :: Python :: 3", 
        "Programming Language :: Python :: 3.7",
    ],
  
)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Aug 14 16:16:09 2021

@author: af1tang
"""
from transformers import GPT2LMHeadModel, GPT2Tokenizer
from convogym._configs import opts
from convogym.utils._device import to_device

try: 
    print("*"*50)
    print("Loading decoder from pretrained")
    model = GPT2LMHeadModel.from_pretrained(opts.model_path)
    print("*"*50)
except Exception as e:
    print(e)
    print("*"*50)
    print("Downloading decoder... ")
    print("*"*50)
    # download decoder
    model = GPT2LMHeadModel.from_pretrained(opts.download_name)
    # save locally
    model.save_pretrained(opts.model_path)
    
model = to_device(model)

try: 
    print("*"*50)
    print("Load tokenizer from pretrained")
    tokenizer = GPT2Tokenizer.from_pretrained(opts.model_path, 
                                            pad_token='<|endoftext|>', cls_token='<|cls|>',
                                            sep_token='<|sep|>')
    print("*"*50)
except Exception as e:
    print(e)
    print("*"*50)
    print("Downloading tokenizer ... ")
    print("*"*50)
    # download personaGPT
    tokenizer = GPT2Tokenizer.from_pretrained(opts.download_name, 
                                        pad_token='<|endoftext|>', cls_token='<|cls|>',
                                        sep_token='<|sep|>')
    # save to local
    tokenizer.save_pretrained(opts.model_path)

# special tokens    
p1_tok, p2_tok, start_tok = tokenizer.encode('<|p1|>')[0], tokenizer.encode('<|p2|>')[0], tokenizer.encode('<|start|>')[0]

# action token
act_tok = tokenizer.encode('<|act|>')[0]

